"""
@file
@brief :epkg:`numpy` functions implemented with :epkg:`onnx`.

.. versionadded:: 0.6

.. versionchanged:: 0.7
"""
import warnings
import numpy
from onnx import onnx_pb as onnx_proto  # pylint: disable=E1101
from onnx.helper import make_tensor
from .onnx_variable import OnnxVar, MultiOnnxVar as xtuple
from .xop import loadop
from .numpy_onnx_impl_body import if_then_else, OnnxVarGraph


def abs(x):
    "See :func:`numpy.abs`."
    OnnxAbs = loadop('Abs')
    return OnnxVar(x, op=OnnxAbs)


def acos(x):
    "See :func:`numpy.acos`."
    OnnxAcos = loadop('Acos')
    return OnnxVar(x, op=OnnxAcos)


def acosh(x):
    "See :func:`numpy.acosh`."
    OnnxAcosh = loadop('Acosh')
    return OnnxVar(x, op=OnnxAcosh)


def amax(x, axis=None, keepdims=0):
    "See :func:`numpy.amax`."
    OnnxReduceMax = loadop('ReduceMax')
    if axis is None:
        return OnnxVar(x, op=OnnxReduceMax, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceMax, keepdims=keepdims, axes=axis)


def amin(x, axis=None, keepdims=0):
    "See :func:`numpy.amin`."
    OnnxReduceMin = loadop('ReduceMin')
    if axis is None:
        return OnnxVar(x, op=OnnxReduceMin, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceMin, keepdims=keepdims, axes=axis)


def arange(start, stop, step=1):
    "See :func:`numpy.arange`, *start*, *stop* must be specified."
    if not isinstance(step, (int, numpy.int64)):
        raise TypeError(  # pragma: no cover
            f"step must be an integer not {type(step)!r}.")
    if isinstance(start, (int, numpy.int64, numpy.int32)):
        start = numpy.array([start], dtype=numpy.int64)
        zero = start == 0
    else:
        zero = False
    if isinstance(stop, (int, numpy.int64, numpy.int32)):
        stop = numpy.array([stop], dtype=numpy.int64)
    value = make_tensor(
        "value", onnx_proto.TensorProto.INT64, (1, ), [step])  # pylint: disable=E1101

    OnnxAdd, OnnxCumSum, OnnxConstantOfShape, OnnxSub = loadop(
        'Add', 'CumSum', 'ConstantOfShape', 'Sub')
    if isinstance(step, (int, numpy.int64, numpy.int32)) and step == 1:
        if zero:
            shape = stop
        else:
            shape = stop - start
        if isinstance(shape, OnnxVar):
            shape = shape.reshape(numpy.array([-1], dtype=numpy.int64))
        _cst = OnnxVar(shape, op=OnnxConstantOfShape, value=value)
        cs = OnnxVar(_cst, numpy.array([0], dtype=numpy.int64),
                     op=OnnxCumSum)
        diff = start - numpy.array([step], dtype=numpy.int64)
        return OnnxVar(cs, diff, op=OnnxAdd)

    if isinstance(step, (int, numpy.int64, numpy.int32)):
        step = numpy.array([step], dtype=numpy.int64)
        if zero:
            shape = stop // step
        else:
            shape = (stop - start) // step
        if isinstance(shape, OnnxVar):
            shape = shape.reshape(numpy.array([-1], dtype=numpy.int64))
        _cst = OnnxVar(shape, op=OnnxConstantOfShape, value=value)
    else:
        # csm = OnnxVar(_cst, step, op=OnnxMul)
        raise NotImplementedError(  # pragma: no cover
            "Not yet implemented.")

    cs = OnnxVar(_cst, numpy.array([0], dtype=numpy.int64),
                 op=OnnxCumSum)
    add = OnnxVar(cs, start, op=OnnxAdd)
    return OnnxVar(add, step, op=OnnxSub)


def argmax(x, axis=0, keepdims=0):
    """
    See :func:`numpy.argmax`.

    .. warning::
        ONNX does not implement default value axis=None.
    """
    if axis is None:
        raise NotImplementedError(  # pragma: no cover
            "ONNX does not allow axis=None.")
    OnnxArgMax = loadop('ArgMax')
    return OnnxVar(x, op=OnnxArgMax, axis=axis, keepdims=keepdims)


def argmin(x, axis=0, keepdims=0):
    """
    See :func:`numpy.argmin`.

    .. warning::
        ONNX does not implement default value axis=None.
    """
    if axis is None:
        raise NotImplementedError(  # pragma: no cover
            "ONNX does not allow axis=None.")
    OnnxArgMin = loadop('ArgMin')
    return OnnxVar(x, op=OnnxArgMin, axis=axis, keepdims=keepdims)


def asin(x):
    "See :func:`numpy.asin`."
    OnnxAsin = loadop('Asin')
    return OnnxVar(x, op=OnnxAsin)


def asinh(x):
    "See :func:`numpy.asinh`."
    OnnxAsinh = loadop('Asinh')
    return OnnxVar(x, op=OnnxAsinh)


def atan(x):
    "See :func:`numpy.atan`."
    OnnxAtan = loadop('Atan')
    return OnnxVar(x, op=OnnxAtan)


def atanh(x):
    "See :func:`numpy.atanh`."
    OnnxAtanh = loadop('Atanh')
    return OnnxVar(x, op=OnnxAtanh)


def ceil(x):
    "See :func:`numpy.ceil`."
    OnnxCeil = loadop('Ceil')
    return OnnxVar(x, op=OnnxCeil)


def clip(x, a_min=None, a_max=None):
    "See :func:`numpy.clip`."
    args = [x]
    if a_min is not None:
        args.append(a_min)
    if a_max is not None:
        args.append(a_max)
    OnnxClip = loadop('Clip')
    return OnnxVar(*args, op=OnnxClip)


def compress(condition, x, axis=None):
    """
    See :func:`numpy.compress`.
    `numpy.compress(condition, x)` or `npnx.compress(x, condition)`.
    """
    OnnxCompress = loadop('Compress')
    if axis is None:
        return OnnxVar(x, condition, op=OnnxCompress)
    return OnnxVar(x, condition, op=OnnxCompress, axis=axis)


def cos(x):
    "See :func:`numpy.cos`."
    OnnxCos = loadop('Cos')
    return OnnxVar(x, op=OnnxCos)


def cosh(x):
    "See :func:`numpy.cosh`."
    OnnxCosh = loadop('Cosh')
    return OnnxVar(x, op=OnnxCosh)


def concat(*x, axis=0):
    """
    Operator concat, handle :func:`numpy.vstack` and
    :func:`numpy.hstack`.
    """
    OnnxConcat = loadop('Concat')
    if len(x) <= 1:
        raise RuntimeError(  # pragma: no cover
            f"N={len(x)}<=1 elements to concatenate.")
    return OnnxVar(*x, op=OnnxConcat, axis=axis)


def cumsum(x, axis):
    "See :func:`numpy.cumsum`."
    OnnxCumSum = loadop('CumSum')
    return OnnxVar(x, axis, op=OnnxCumSum)


def cst(x, dtype=None):
    """
    Creates a constant. `log(x) + numpy.float32(1)` works
    but `numpy.float32(32) + log(x)` fails because Python
    calls `numpy.float32.__add__` instead of
    `OnnxVar.__add__`. With this function, expression
    `cst(1.) + log(x)` is valid. Parameter `dtype` is
    used to overwrite the default dtype (`numpy.float32`
    for floats and `numpy.int64` for ints.
    """
    OnnxIdentity = loadop('Identity')
    if isinstance(x, float):
        return OnnxVar(numpy.array([x], dtype=dtype or numpy.float32),
                       op=OnnxIdentity)
    if isinstance(x, int):
        return OnnxVar(numpy.array([x], dtype=dtype or numpy.int64),
                       op=OnnxIdentity)
    if isinstance(x, numpy.ndarray):
        return OnnxVar(x, op=OnnxIdentity)
    if hasattr(x, 'dtype'):
        if dtype is not None:
            raise RuntimeError(  # pragma: no cover
                f"dtype is not used because x is of type {type(x)!r}.")
        return OnnxVar(numpy.array([x], dtype=x.dtype),
                       op=OnnxIdentity)
    raise NotImplementedError(  # pragma: no cover
        f"Unable to convert type {type(x)!r} into a constant.")


def det(x):
    "See :func:`numpy.linalg:det`."
    OnnxDet = loadop('Det')
    return OnnxVar(x, op=OnnxDet)


def dot(a, b):
    "See :func:`numpy.dot`"
    warnings.warn(
        "npnx.dot is equivalent to npnx.matmul == numpy.matmul "
        "!= numpy.dot with arrays with more than 3D dimensions.")
    OnnxMatMul = loadop('MatMul')
    return OnnxVar(a, b, op=OnnxMatMul)


def matmul(a, b):
    "See :func:`numpy.matmul`."
    OnnxMatMul = loadop('MatMul')
    return OnnxVar(a, b, op=OnnxMatMul)


def einsum(*x, equation=None):
    "See :func:`numpy.einsum`."
    OnnxEinsum = loadop('Einsum')
    return OnnxVar(*x, op=OnnxEinsum, equation=equation)


def erf(x):
    "See :epkg:`scipy:special:erf`."
    OnnxErf = loadop('Erf')
    return OnnxVar(x, op=OnnxErf)


def exp(x):
    "See :func:`numpy.exp`."
    OnnxExp = loadop('Exp')
    return OnnxVar(x, op=OnnxExp)


def expand_dims(x, axis):
    "See :func:`numpy.expand_dims`."
    if not isinstance(axis, int):
        raise NotImplementedError(  # pragma: no cover
            f"This function only allows integer for axis not {type(axis)!r}.")
    OnnxUnsqueeze = loadop('Unsqueeze')
    return OnnxVar(x, numpy.array([axis], dtype=numpy.int64),
                   op=OnnxUnsqueeze)


def expit(x):
    "See :epkg:`scipy:special:expit`."
    OnnxSigmoid = loadop('Sigmoid')
    return OnnxVar(x, op=OnnxSigmoid)


def floor(x):
    "See :func:`numpy.floor`."
    OnnxFloor = loadop('Floor')
    return OnnxVar(x, op=OnnxFloor)


def hstack(*x):
    "See :func:`numpy.hstack`."
    if len(x) <= 1:
        raise RuntimeError(  # pragma: no cover
            f"N={len(x)}<=1 elements to concatenate.")
    OnnxConcat = loadop('Concat')
    return OnnxVar(*x, op=OnnxConcat, axis=-1)


def isnan(x):
    "See :func:`numpy.isnan`."
    OnnxIsNaN = loadop('IsNaN')
    return OnnxVar(x, op=OnnxIsNaN)


def identity(x):
    "Identity."
    OnnxIdentity = loadop('Identity')
    return OnnxVar(x, op=OnnxIdentity)


def log(x):
    "See :func:`numpy.log`."
    OnnxLog = loadop('Log')
    return OnnxVar(x, op=OnnxLog)


def log1p(x):
    "See :func:`numpy.log1p`."
    OnnxLog, OnnxAdd = loadop('Log', 'Add')
    x1 = OnnxVar(x, numpy.array([1], dtype=x.dtype),
                 op=OnnxAdd)
    return OnnxVar(x1, op=OnnxLog)


def mean(x, axis=None, keepdims=0):
    "See :func:`numpy.mean`."
    OnnxReduceMean = loadop('ReduceMean')
    if axis is None:
        return OnnxVar(x, op=OnnxReduceMean, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceMean, keepdims=keepdims, axes=axis)


def onnx_if(condition, then_branch, else_branch):
    """
    Implements a test with onnx syntax.

    :param condition: condition (@see cl OnnxVar)
    :param then_branch: then branch, of type @see cl if_then_else
    :param else_branch: else branch, of type @see cl if_then_else
    :return: result (@see cl OnnxVar)
    """
    OnnxIf = loadop('If')
    if isinstance(then_branch, numpy.ndarray):
        then_branch = if_then_else(then_branch)
    if not isinstance(then_branch, if_then_else):
        raise TypeError(
            "Parameter then_branch is not of type "
            "'if_then_else' but %r." % type(then_branch))
    if isinstance(else_branch, numpy.ndarray):
        else_branch = if_then_else(else_branch)
    if not isinstance(else_branch, if_then_else):
        raise TypeError(
            "Parameter then_branch is not of type "
            "'if_then_else' but %r." % type(else_branch))
    return OnnxVarGraph(
        condition, then_branch=then_branch,
        else_branch=else_branch, op=OnnxIf)


def pad(x, pads, constant_value=None, mode='constant'):
    """
    It does not implement :func:`numpy.pad` but the ONNX version
    :func:`onnx_pad <mlprodict.onnxrt.ops_cpu.op_pad.onnx_pad>`.
    """
    OnnxPad = loadop(('', 'Pad'))
    if constant_value is None:
        return OnnxVar(x, pads, op=OnnxPad, mode=mode)
    return OnnxVar(x, pads, constant_value, op=OnnxPad, mode=mode)


def prod(x, axis=None, keepdims=0):
    "See :func:`numpy.prod`."
    OnnxReduceProd = loadop('ReduceProd')
    if axis is None:
        return OnnxVar(x, op=OnnxReduceProd, keepdims=keepdims)
    if not isinstance(axis, list):
        axis = [axis]
    return OnnxVar(x, op=OnnxReduceProd, keepdims=keepdims, axes=axis)


def relu(x):
    "relu"
    OnnxRelu = loadop('Relu')
    return OnnxVar(x, op=OnnxRelu)


def reciprocal(x):
    "See :func:`numpy.reciprocal`."
    OnnxReciprocal = loadop('Reciprocal')
    return OnnxVar(x, op=OnnxReciprocal)


def round(x):
    "See :func:`numpy.round`."
    OnnxRound = loadop('Round')
    return OnnxVar(x, op=OnnxRound)


def sigmoid(x):
    "See :epkg:`scipy:special:expit`."
    OnnxSigmoid = loadop('Sigmoid')
    return OnnxVar(x, op=OnnxSigmoid)


def sign(x):
    "See :func:`numpy.sign`."
    OnnxSign = loadop('Sign')
    return OnnxVar(x, op=OnnxSign)


def sin(x):
    "See :func:`numpy.sin`."
    OnnxSin = loadop('Sin')
    return OnnxVar(x, op=OnnxSin)


def sinh(x):
    "See :func:`numpy.sinh`."
    OnnxSinh = loadop('Sinh')
    return OnnxVar(x, op=OnnxSinh)


def sqrt(x):
    "See :func:`numpy.sqrt`."
    OnnxSqrt = loadop('Sqrt')
    return OnnxVar(x, op=OnnxSqrt)


def squeeze(x, axis=None):
    "See :func:`numpy.squeeze`."
    OnnxSqueeze = loadop('Squeeze')
    if axis is None:
        raise NotImplementedError(  # pragma: no cover
            "The case where all empty dimensions are removed is not "
            "implemented.")
    if isinstance(axis, int):
        raise RuntimeError(  # pragma: no cover
            "axis must be a tensor.")
    return OnnxVar(x, axis, op=OnnxSqueeze)


def sum(x, axis=None, keepdims=0):
    "See :func:`numpy.sum`."
    OnnxReduceSum = loadop('ReduceSum')
    if axis is None:
        return OnnxVar(x, op=OnnxReduceSum, keepdims=keepdims)
    return OnnxVar(x, numpy.array([axis], dtype=numpy.int64),
                   op=OnnxReduceSum, keepdims=keepdims)


def tan(x):
    "See :func:`numpy.tan`."
    OnnxTan = loadop('Tan')
    return OnnxVar(x, op=OnnxTan)


def tanh(x):
    "See :func:`numpy.tanh`."
    OnnxTanh = loadop('Tanh')
    return OnnxVar(x, op=OnnxTanh)


def topk(x, k, axis=-1, largest=1, sorted=1):
    "See :func:`numpy.argsort`."
    OnnxTopK = loadop('TopK')
    return xtuple(x, k, op=OnnxTopK, axis=axis, largest=largest,
                  sorted=sorted)


def transpose(x, perm=(1, 0)):
    "See :func:`numpy.transpose`."
    OnnxTranspose = loadop('Transpose')
    return OnnxVar(x, op=OnnxTranspose, perm=list(perm))


def unsqueeze(x, axes):
    "See :func:`numpy.expand_dims`."
    OnnxUnsqueeze = loadop('Unsqueeze')
    if isinstance(axes, int):
        axes = numpy.array([axes], dtype=numpy.int64)
    return OnnxVar(x, axes, op=OnnxUnsqueeze)


def vstack(*x):
    "See :func:`numpy.vstack`."
    OnnxConcat = loadop('Concat')
    if len(x) <= 1:
        raise RuntimeError(  # pragma: no cover
            f"N={len(x)}<=1 elements to concatenate.")
    return OnnxVar(*x, op=OnnxConcat, axis=0)


def where(cond, x, y):
    "See :func:`numpy.where`."
    OnnxWhere = loadop('Where')
    return OnnxVar(cond, x, y, op=OnnxWhere)



"""
@file
@brief Investigate issues happening with float32.
"""
from io import BytesIO
import numpy
from numpy.random import randint
from onnx import ModelProto, FunctionProto, GraphProto, load
from onnx.checker import check_model


class MissingInputError(RuntimeError):
    "Raised when an input is missing."
    pass


def astype_range(arr, dtype=numpy.float32, force=1):
    """
    Computes ranges for every number in an array
    once converted into *float32*. The function returns
    two matrices which produces two numbers
    *a* et *b*, the number rounded to float32
    is in interval :math:`[a, b]`.

    @param      arr     array
    @param      dtype   type to convert to
    @param      force   does something like *[i] +/- force |i - [i]|*
    @return             minimum, maximum
    """
    conv = arr.astype(dtype)
    delta = numpy.abs(arr - conv)
    delta = numpy.maximum(numpy.abs(arr) * 1e-7, delta)
    maxa = (conv + delta * force).astype(dtype)
    mina = (conv - delta * force).astype(dtype)
    return mina, maxa


def enumerate_random_inputs(inputs, n=100, dtype=numpy.float32, force=1):
    """
    Enumerates random matrices.

    @param      inputs      inputs (dictionary)
    @param      n           number of iterations
    @param      dtype       type to convert to
    @param      force       does something like *[i] +/- force |i - [i]|*
    """
    keys = list(inputs)
    ranges = {k: astype_range(v, dtype=dtype, force=force)
              for k, v in inputs.items()}
    for _ in range(n):
        new_inputs = {}
        for k in keys:
            rnd = randint(0, 2, inputs[k].size).reshape(  # pylint: disable=E1101
                inputs[k].shape)  # pylint: disable=E1101
            if rnd.min() == rnd.max() or rnd.max() != 1:
                raise RuntimeError(  # pragma: no cover
                    "Minimum and maximum are equal or maximum is not 1. "
                    "Randomness failed.")
            rnd = rnd.astype(dtype)
            ma1 = ranges[k][0] * rnd
            ma2 = ranges[k][1] * (-(rnd - 1))
            inp = (ma1 + ma2)
            new_inputs[k] = inp
        yield new_inputs


def onnx_shaker(oinf, inputs, output_fct, n=100, dtype=numpy.float32, force=1):
    """
    Shakes a model :epkg:`ONNX`.
    Explores the ranges for every prediction.
    Uses @see fn astype_range

    @param      oinf        object of type @see cl OnnxInference
    @param      inputs      inputs
    @param      output_fct  output function which extracts
                            a single array from the output
    @param      dtype       type to convert to
    @param      force       does something like *[i] +/- force |i - [i]|*
    @return                 ranges for each predictions

    See notebook :ref:`onnxshakerrst` for an example of use.
    """
    results = None
    for i, new_inputs in enumerate(enumerate_random_inputs(
            inputs, n=n, dtype=dtype, force=force)):
        res_ = oinf.run(new_inputs)
        res = output_fct(res_)
        sq = numpy.squeeze(res)
        if len(sq.shape) != 1:
            raise ValueError(  # pragma: no cover
                f"The function only works with shape={sq.shape}")
        if results is None:
            results = numpy.empty((sq.shape[0], n), dtype=sq.dtype)
        results[:, i] = sq

    results.sort(axis=1)
    return results


def check_onnx(model, use_onnx=False, known_results=None,
               path=None):
    """
    Checks consistency of the model.

    :param model: onnx graph
    :param use_onnx: calls `onnx.checker.check_model`
    :param known_results: known results
    :param path: path to a node (through subgraphs)
    """
    if isinstance(model, bytes):
        model = load(BytesIO(model))

    def raise_missing(name, node, p, kn):
        raise MissingInputError(
            "Missing input %r in node type=%r and name=%r "
            "path=%r, known=\n%s\n--ONNX--\n%s" % (
                name, node.op_type, node.name,
                [n.name for n in p], "\n".join(sorted(kn)),
                str(model)))

    if isinstance(model, ModelProto):
        try:
            check_onnx(model.graph, known_results=known_results)
        except MissingInputError as e:
            raise MissingInputError(
                f"Wrong ONNX model\n--ONNX\n{str(model)}") from e
        for f in model.functions:
            check_onnx(f)
        return
    if known_results is None:
        known_results = {}
    else:
        known_results = known_results.copy()
    if isinstance(model, FunctionProto):
        for i in model.input:
            known_results[i] = i
    elif isinstance(model, GraphProto):
        for i in model.input:
            known_results[i.name] = i
        for i in model.initializer:
            known_results[i.name] = i
    else:
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(model)!r}.")

    if path is None:
        path = []
    else:
        path = path.copy()

    for node in model.node:
        for i in node.input:
            if i == '':
                # optional input
                continue
            if i not in known_results:
                raise_missing(i, node, path + [node], known_results)
            for att in node.attribute:
                if hasattr(att, 'g') and att.g is not None:
                    check_onnx(att.g, use_onnx=use_onnx,
                               known_results=known_results,
                               path=path + [att, node])
        for o in node.output:
            known_results[o] = node

    if use_onnx:
        check_model(model)



"""
@file
@brief Direct calls to libraries :epkg:`BLAS` and :epkg:`LAPACK`.
The wrapper for GEMM still does not work for matrices
which are not square.
"""
from libc.stdio cimport printf

import numpy
cimport numpy
cimport cython
numpy.import_array()
# cimport scipy.linalg.cython_lapack as cython_lapack
cimport scipy.linalg.cython_blas as cython_blas


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void c_dgemm_dot(const double* pa, const double* pb,
                      int M, int N, int K,
                      int transA, int transB,
                      double* pc) nogil:
    """
    Wrapper for gemm.    
    """

    cdef:
        const char * cst = "nt"
        int lda = K if transA else M
        int ldb = K if transB else N
        int ldc = M
        double alpha = 1.
        double beta = 0.

    cython_blas.dgemm(&cst[transA], &cst[transB],
                      &M, &N, &K,
                      &alpha, pa, &lda, pb, &ldb,
                      &beta, pc, &ldc)



@cython.boundscheck(False)
@cython.wraparound(False)
def dgemm_dot(double [:, ::1] A, double [:, ::1] B, transA, transB, double [:, ::1] C):
    """
    Wrapper for gemm.    
    """
    
    cdef:
        int ta = 1 if transA else 0
        int tb = 1 if transB else 0
        int M = A.shape[1 - ta]
        int N = B.shape[tb]
        int K = A.shape[ta]
    
    c_dgemm_dot(&A[0, 0], &B[0, 0], M, N, K, ta, tb, &C[0, 0])


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void c_sgemm_dot(const float* pa, const float* pb,
                      int M, int N, int K,
                      int transA, int transB,
                      float* pc) nogil:
    """
    Wrapper for gemm.    
    """

    cdef:
        const char * cst = "nt"
        int lda = K if transA else M
        int ldb = K if transB else N
        int ldc = M
        float alpha = 1.
        float beta = 0.

    cython_blas.sgemm(&cst[transA], &cst[transB],
                      &M, &N, &K,
                      &alpha, pa, &lda, pb, &ldb,
                      &beta, pc, &ldc)



@cython.boundscheck(False)
@cython.wraparound(False)
def sgemm_dot(float [:, ::1] A, float [:, ::1] B, transA, transB, float [:, ::1] C):
    """
    Wrapper for gemm.    
    """

    cdef:
        int ta = 1 if transA else 0
        int tb = 1 if transB else 0
        int M = A.shape[1 - ta]
        int N = B.shape[tb]
        int K = A.shape[ta]

    c_sgemm_dot(&A[0, 0], &B[0, 0], M, N, K, ta, tb, &C[0, 0])



"""
@file
@brief Converters for models from :epkg:`mlinsights`.
"""
from sklearn.base import is_classifier
from skl2onnx import get_model_alias
from skl2onnx.common._registration import (
    get_shape_calculator, _converter_pool, _shape_calculator_pool)
from skl2onnx._parse import _parse_sklearn
from skl2onnx.common._apply_operation import apply_identity
from skl2onnx.common._topology import Scope, Variable  # pylint: disable=E0611,E0001
from skl2onnx._supported_operators import sklearn_operator_name_map


def parser_transfer_transformer(scope, model, inputs, custom_parsers=None):
    """
    Parser for :epkg:`TransferTransformer`.
    """
    if len(inputs) != 1:
        raise RuntimeError(  # pragma: no cover
            "Only one input (not %d) is allowed for model type %r."
            "" % (len(inputs), type(model)))
    if custom_parsers is not None and model in custom_parsers:
        return custom_parsers[model](
            scope, model, inputs, custom_parsers=custom_parsers)

    if model.method == 'predict_proba':
        name = 'probabilities'
    elif model.method == 'transform':
        name = 'variable'
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to defined the output for method='{}' and model='{}'."
            "".format(model.method, model.__class__.__name__))

    prob = scope.declare_local_variable(name, inputs[0].type.__class__())
    alias = get_model_alias(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    this_operator.inputs = inputs
    this_operator.outputs.append(prob)
    return this_operator.outputs


def shape_calculator_transfer_transformer(operator):
    """
    Shape calculator for :epkg:`TransferTransformer`.
    """
    if len(operator.inputs) != 1:
        raise RuntimeError(  # pragma: no cover
            "Only one input (not %d) is allowed for model %r."
            "" % (len(operator.inputs), operator))
    op = operator.raw_operator
    alias = get_model_alias(type(op.estimator_))
    calc = get_shape_calculator(alias)

    options = (None if not hasattr(operator.scope, 'options')
               else operator.scope.options)
    if is_classifier(op.estimator_):
        if options is None:
            options = {}
        options = {id(op.estimator_): {'zipmap': False}}
    registered_models = dict(
        conv=_converter_pool, shape=_shape_calculator_pool,
        aliases=sklearn_operator_name_map)
    scope = Scope('temp', options=options,
                  registered_models=registered_models)
    inputs = [
        Variable(v.onnx_name, v.onnx_name, type=v.type, scope=scope)
        for v in operator.inputs]
    res = _parse_sklearn(scope, op.estimator_, inputs)
    this_operator = res[0]._parent
    calc(this_operator)

    if op.method == 'predict_proba':
        operator.outputs[0].type = this_operator.outputs[1].type
    elif op.method == 'transform':
        operator.outputs[0].type = this_operator.outputs[0].type
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to defined the output for method='{}' and model='{}'.".format(
                op.method, op.__class__.__name__))
    if len(operator.inputs) != 1:
        raise RuntimeError(  # pragma: no cover
            "Only one input (not %d) is allowed for model %r."
            "" % (len(operator.inputs), operator))


def convert_transfer_transformer(scope, operator, container):
    """
    Converters for :epkg:`TransferTransformer`.
    """
    op = operator.raw_operator

    opts = scope.get_options(op)
    if opts is None:
        opts = {}
    if is_classifier(op.estimator_):
        opts['zipmap'] = False
    container.add_options(id(op.estimator_), opts)
    scope.add_options(id(op.estimator_), opts)

    outputs = _parse_sklearn(scope, op.estimator_, operator.inputs)

    if op.method == 'predict_proba':
        index = 1
    elif op.method == 'transform':
        index = 0
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to defined the output for method='{}' and model='{}'."
            "".format(op.method, op.__class__.__name__))

    apply_identity(scope, outputs[index].onnx_name,
                   operator.outputs[0].full_name, container,
                   operator_name=scope.get_unique_operator_name("IdentityTT"))



"""
@file
@brief One class which visits a syntax tree.
"""
import pprint
import numpy


class CodeTranslator:
    """
    Class which converts a Python function into
    something else. It must implements
    methods *visit* and *depart*.
    """

    def __init__(self, visitor):
        """
        :param visitor: :class:`CodeNodeVisitor
            <mlprodict.onnx_tools.onnx_grammar.node_visitor_translator>`
        """
        self._visitor = visitor

    def export(self, context=None, **kwargs):
        """
        Exports the parsed :epkg:`python` code
        into something.
        """
        raise NotImplementedError(  # pragma: no cover
            "This function should be overwritten.")

    def visit(self, node, info):
        """
        Visits a node.

        @param      node        visited node
        @param      info        info extracted by the visitor
        """
        raise NotImplementedError(  # pragma: no cover
            "This function should be overwritten.")

    def depart(self, node, info):
        """
        Leaves a node.

        @param      node        visited node
        @param      info        info extracted by the visitor
        """
        raise NotImplementedError(  # pragma: no cover
            "This function should be overwritten.")


class OnnxTranslator(CodeTranslator):
    """
    Class which converts a Python function into
    an :epkg:`ONNX` function. It must implements
    methods *visit* and *depart*.
    """
    _binary_operators = {
        'Add': 'Add', 'Div': 'Div',
        'Mult': 'Mul', 'Sub': 'Sub',
        'Pow': 'Pow', 'MatMult': 'MatMul',
    }

    _unary_operators = {
        'Sub': 'Neg',
    }

    _numpy2onnx_op = {
        'absolute': 'Abs',
        'cos': 'Cos',
        'exp': 'Exp',
        'power': 'Pow',
        'transpose': 'Transpose',
        'sin': 'Sin',
        # complex function
        'inner': 'inner',
    }

    _parameter_mapping = {
        'Transpose': {'axes': 'perm'}
    }

    class Parameter:
        """
        Holds parameter information.
        """

        def __init__(self, name, value=('#NODEFAULT#', ), annotation=None):
            """
            @param      name        parameter name
            @param      value       parameter value
            """
            self.name = name
            self.value = value
            self.annotation = annotation

        @staticmethod
        def format_value(value):
            """
            Returns a formatted value in python code.
            """
            if isinstance(value, str):
                return '"{}"'.format(value.replace('"', '\\"').replace('\\', '\\\\'))
            if isinstance(value, list):
                return f"[{', '.join(map(OnnxTranslator.Parameter.format_value, value))}]"
            if isinstance(value, tuple):
                if value == ('#NODEFAULT#', ):
                    return None
                return f"({', '.join(map(OnnxTranslator.Parameter.format_value, value))})"
            return str(value)

        @property
        def formatted_value(self):
            """
            Returns a formatted value in python code.
            """
            return OnnxTranslator.Parameter.format_value(self.value)

        def __str__(self):
            """
            Into python syntax.
            """
            rows = [self.name]
            if self.value != ('#NODEFAULT#', ):
                rows.append('=')
                rows.append(self.formatted_value)
            return ''.join(rows)

    def __init__(self, visitor):
        """
        :param visitor: :class:`CodeNodeVisitor
            <mlprodict.onnx_tools.onnx_grammar.node_visitor_translator>`
        """
        CodeTranslator.__init__(self, visitor)
        self._stack = []
        self._code_fct = None

    def _is_stacked(self, name):
        for line in self._stack:
            if line[0] == name:
                return True
        return False

    def _get_last(self, name, info=None):
        if len(self._stack) == 0:
            raise RuntimeError("Stack is empty.")  # pragma: no cover
        last = self._stack[-1]
        if ((isinstance(name, str) and last[0] != name) or
                (isinstance(name, tuple) and last[0] not in name)):
            raise RuntimeError(  # pragma: no cover
                "Last item is not '{}'\n{}\n---\n{}".format(
                    name, pprint.pformat(self._stack),
                    pprint.pformat(info) if info else ""))
        return last

    def make_msg(self, info):
        """
        Make a message with line and column information.
        """
        lineno = '?'
        col_offset = '?'
        if isinstance(info, dict):
            if 'node' in info:
                node = info['node']
                lineno = node.lineno
                col_offset = node.col_offset
            else:
                if 'lineno' in info:
                    lineno = info['lineno']
                if 'col_offset' in info:
                    col_offset = info['col_offset']
        else:
            if hasattr(info, 'lineno'):
                lineno = info.lineno
            if hasattr(info, 'col_offset'):
                col_offset = info.col_offset

        return f"line {lineno}, col {col_offset}"

    def export(self, context=None, format='code',  # pylint: disable=W0221
               output_names=None):
        """
        Returns an :epkg:`ONNX` graph or a piece
        of code which could generate the graph.

        @param      context         function used in the function code
        @param      format          ``'code'``
        @param      output_names    add code in the final function
                                    to overwrite the names of the
                                    outputs in the :epkg:`ONNX` graph
        @return                     string or :epkg:`onnx` graph

        This method is used in function @see fn translate_fct2onnx.
        An example of code can be found there.
        """
        if self._code_fct is None:
            raise RuntimeError(  # pragma: no cover
                "No python code was parsed.")
        if context is None:
            context = {}

        def find_onnx_correspondance(fct, info):
            if isinstance(fct, numpy.ufunc):
                name = fct.__name__
            elif callable(fct) and getattr(fct, '__module__', '') in (
                    'numpy', 'numpy.core.fromnumeric'):
                name = fct.__name__
            elif callable(fct) and fct.__name__.startswith("py_"):
                return fct
            else:
                name = None
            if name is not None and name not in OnnxTranslator._numpy2onnx_op:
                raise RuntimeError(  # pragma: no cover
                    "Unable to find a correspondance to '{}' at {} in \n{}".format(
                        name, self.make_msg(info),
                        "\n".join(sorted(OnnxTranslator._numpy2onnx_op))))
            if name is not None:
                return OnnxTranslator._numpy2onnx_op[name]
            if isinstance(fct, str):
                return fct
            raise RuntimeError(  # pragma: no cover
                "Unable to find a correspondance for function name '{}' in module '{}', "
                "'{}' (type {}) at {}.".format(
                    name, getattr(fct, '__module__', ''),
                    fct, type(fct), self.make_msg(info)))

        def write_expression(stack_fct_used, expr, indent, parameter_mapping=None):
            if isinstance(expr, str):
                # an argument
                return [f"{' ' * indent * 4}{expr}"]
            if isinstance(expr, (int, float)):
                # an argument
                return [f"{' ' * indent * 4}{expr}"]
            if isinstance(expr, OnnxTranslator.Parameter):
                if parameter_mapping is None:
                    name = expr.name
                else:
                    name = parameter_mapping.get(expr.name, expr.name)
                return [f"{' ' * indent * 4}{name}={expr.formatted_value}"]
            rows = []
            if isinstance(expr, tuple):
                expr = [expr]
            for op, args in expr:
                if op == 'BinOp':
                    opname = args["op"]
                    opon = args["args"]
                    onnx_name = OnnxTranslator._binary_operators[opname]
                    rows.append(
                        f"{' ' * indent * 4}Onnx{onnx_name}(")
                    for expr2 in opon:
                        sexpr2 = write_expression(
                            stack_fct_used, expr2, indent + 1)
                        if any(filter(lambda s: 'op_version="op_version"' in s, sexpr2)):
                            continue  # pragma: no cover
                        rows.extend(sexpr2)
                        rows[-1] += ","
                    rows.append(
                        f"{' ' * (indent + 1) * 4}op_version=op_version")
                    rows.append(f"{' ' * indent * 4})")
                elif op == 'UnaryOp':
                    opname = args["op"]
                    opon = args["args"]
                    onnx_name = OnnxTranslator._unary_operators[opname]
                    rows.append(
                        f"{' ' * indent * 4}Onnx{onnx_name}(")
                    for expr2 in opon:
                        sexpr2 = write_expression(
                            stack_fct_used, expr2, indent + 1)
                        if any(filter(lambda s: 'op_version="op_version"' in s, sexpr2)):
                            continue
                        rows.extend(sexpr2)
                        rows[-1] += ","
                    rows.append(
                        f"{' ' * (indent + 1) * 4}op_version=op_version")
                    rows.append(f"{' ' * indent * 4})")
                elif op == 'Call':
                    name = args['name']
                    if name.startswith("onnx_"):
                        raise RuntimeError("The code must not use a function prefixed by 'onnx_' (%s). "
                                           "It indicates that function manipulate ONNX node and "
                                           "the fonction to convert must only deal with arrays." % name)
                    if name not in context:
                        raise RuntimeError(
                            "Unable to find function '{}' at {} in context\n{}\n--\n{}".format(
                                name, self.make_msg(args),
                                '\n'.join(sorted(context)),
                                pprint.pformat(args)))
                    op_conv = find_onnx_correspondance(context[name], args)
                    if callable(op_conv) and op_conv.__name__.startswith('py_'):
                        rows.append(
                            f"{' ' * indent * 4}{op_conv.__name__}(")
                    elif callable(op_conv) and op_conv.__name__.startswith('onnx_'):
                        stack_fct_used.append(op_conv.__name__)
                        rows.append(
                            f"{' ' * indent * 4}{op_conv}(")
                    else:
                        prefix = "onnx_" if 'a' <= op_conv[0] <= 'z' else 'Onnx'
                        if prefix == "onnx_":
                            stack_fct_used.append(
                                f"{prefix}{op_conv}")
                            prefix = '_' + prefix
                        rows.append(
                            f"{' ' * indent * 4}{prefix}{op_conv}(")

                    opon = args["args"]
                    opon = opon[1:]
                    for expr2 in opon:
                        sexpr2 = write_expression(
                            stack_fct_used, expr2, indent + 1,
                            OnnxTranslator._parameter_mapping.get(op_conv, None))
                        if any(filter(lambda s: 'op_version="op_version"' in s, sexpr2)):
                            continue
                        rows.extend(sexpr2)
                        rows[-1] += ","
                    rows.append(
                        f"{' ' * (indent + 1) * 4}op_version=op_version")
                    rows.append(f"{' ' * indent * 4})")
                else:
                    raise RuntimeError(  # pragma: no cover
                        f"Unable to interpret '{expr}'.")
            return rows

        def write_function(stack_fct_used, to_replaces, node):
            rows = []
            name, args = node
            if name != 'FunctionDef':
                raise RuntimeError(  # pragma: no cover
                    "The code being translated should be a single function not "
                    "'{}' at {}.".format(name, self.make_msg(args)))
            list_args = list(map(str, args['args']))
            if all(map(lambda s: 'dtype=' not in s, list_args)):
                list_args.append("dtype=numpy.float32")
            if all(map(lambda s: 'op_version=' not in s, list_args)):
                list_args.append("op_version=None")
            fct_name = args['name']
            rows.append(f"def {fct_name}({', '.join(list_args)}):")
            indent = 1

            to_replace = f"# __HEADER__{id(node)}"
            to_replaces.append(to_replace)
            rows.append(f"{' ' * (indent * 4)}{to_replace}")

            code = args['code']
            for op, args in code:
                if op == "Assign":
                    name = args['name']
                    args = args["args"]
                    rows.append(f"{' ' * (indent * 4)}{name} = (")
                    rows.extend(write_expression(
                        stack_fct_used, args, indent + 1))
                    rows.append(f"{' ' * (indent * 4)})")
                elif op == "Return":
                    args = args["code"]
                    if output_names is None:
                        rows.append(f"{' ' * (indent * 4)}return (")
                        rows.extend(write_expression(
                            stack_fct_used, args, indent + 1))
                        rows.append(f"{' ' * (indent * 4)})")
                    else:
                        rows.append(
                            f"{' ' * (indent * 4)}return OnnxIdentity(")
                        subrows = write_expression(
                            stack_fct_used, args, indent + 1)
                        subrows[-1] += ","
                        rows.extend(subrows)
                        rows.append("{}output_names={},".format(
                            " " * ((indent + 1) * 4), str(output_names)))
                        rows.append(
                            f"{' ' * ((indent + 1) * 4)}op_version=op_version")
                        rows.append(f"{' ' * (indent * 4)})")
                else:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to process operator '{}' at {}. "
                        "Make sure it is either an affectation, "
                        "either a return.".format(op, self.make_msg(args)))
            return rows

        stack_fct_used = []
        to_replaces = []
        rows = write_function(stack_fct_used, to_replaces, self._code_fct)

        # handling dtype parameter
        if len(to_replaces) != 1:
            raise RuntimeError(  # pragma: no cover
                "The following code misses a placeholder:\n{}".format(
                    "\n".join(rows)))
        index = -1
        for i, row in enumerate(rows):
            if to_replaces[0] in row:
                index = i
                break

        header = []
        for fct in stack_fct_used:
            header.append(
                "    _{0} = lambda *args, op_version=op_version, **kwargs: {0}(*args, dtype=dtype, "
                "op_version=op_version, **kwargs)".format(fct))
        if len(header) > 0:
            header.append('')
        rows[index:index + 1] = header

        return "\n".join(rows)

    def visit(self, node, info):
        """
        Visits a node.

        @param      node        visited node
        @param      info        info extracted by the visitor
        """
        if 'type' not in info:
            return

        kind = info['type']
        if kind == "Module":
            return
        if kind == "FunctionDef":
            if self._is_stacked('FunctionDef'):
                raise RuntimeError("Nested functions are not allowed at {}.".format(
                    self.make_msg(node)))
            self._stack.append(
                ('FunctionDef', {'args': [], 'code': [], 'name': info['name'], 'default': [],
                                 'lineno': node.lineno, 'col_offset': node.col_offset}))
            return
        if kind == "arguments":
            _, buf = self._get_last('FunctionDef')
            return
        if kind == "arg":
            return
        if kind == "Assign":
            self._stack.append(
                ('Assign', {'args': [], 'lineno': node.lineno, 'col_offset': node.col_offset}))
            return
        if kind in ('Name', 'Cst'):
            self._get_last(
                ('Assign', 'BinOp', 'Call', 'Return', 'FunctionDef', 'keyword', 'UnaryOp'))
            return
        if kind == 'BinOp':
            self._stack.append(
                ('BinOp', {'args': [], 'lineno': node.lineno, 'col_offset': node.col_offset}))
            return
        if kind == 'UnaryOp':
            self._stack.append(
                ('UnaryOp', {'args': [], 'lineno': node.lineno, 'col_offset': node.col_offset}))
            return
        if kind in OnnxTranslator._binary_operators:
            _, buf = self._get_last(('BinOp', 'UnaryOp'))
            buf['op'] = kind
            return
        if kind == 'Call':
            self._stack.append(
                ('Call', {'name': info['str'], 'args': [], 'lineno': node.lineno,
                          'col_offset': node.col_offset}))
            return
        if kind == 'Return':
            self._get_last('FunctionDef')
            self._stack.append(
                ('Return', {'code': [], 'lineno': node.lineno, 'col_offset': node.col_offset}))
            return
        if kind == "Attribute":
            if info.get('str', '') == 'T':
                raise NotImplementedError(  # pragma: no cover
                    "Transpose should be done with numpy.transpose not with .T'{}' "
                    "at {}\n{}\n---\n{}".format(
                        info.get('type', '?'), self.make_msg(node),
                        pprint.pformat(info), pprint.pformat(self._stack)))
            self._get_last('Call')
            return
        if kind == 'keyword':
            self._get_last('Call')
            self._stack.append(
                ('keyword', {'name': f"{node.arg}",
                             'lineno': getattr(node, 'lineno', '?'),
                             'col_offset': getattr(node, 'col_offset', '?')}))
            return
        if kind == 'List':
            self._get_last('keyword')
            self._stack.append(
                ('List', {'elts': [], 'lineno': getattr(node, 'lineno', '?'),
                          'col_offset': getattr(node, 'col_offset', '?')}))
            return
        if kind == 'Num':
            self._get_last(('List', 'UnaryOp', 'BinOp', 'FunctionDef', 'Call'))
            return
        if kind == 'Str':
            self._get_last('keyword')
            return

        raise NotImplementedError(  # pragma: no cover
            "Unable to interpret kind '{}' at {}\n{}\n---\n{}".format(
                info.get('type', '?'), self.make_msg(
                    node), pprint.pformat(info),
                pprint.pformat(self._stack)))

    def _fix_default_values(self, code_fct):
        """
        Maps default values with parameter names.
        """
        nbdef = len(code_fct[1]['default'])
        nbpar = len(code_fct[1]['args'])
        args = []
        for i in range(nbpar):
            name, annotation = code_fct[1]['args'][i]
            j = nbdef - (nbpar - i)
            if j >= 0:
                default = code_fct[1]['default'][j]
                p = OnnxTranslator.Parameter(
                    name, annotation=annotation, value=default)
            else:
                p = OnnxTranslator.Parameter(name, annotation=annotation)
            args.append(p)
        code_fct[1]['args'] = args

    def _post_process(self, op, node):
        """
        Simplifies some operator such as ``OnnxNeg(2)``.
        """
        if op is None and 'args' in node:
            for i in range(len(node['args'])):
                if not isinstance(node['args'][i], tuple):
                    continue
                o, v = node['args'][i]
                if (o == 'UnaryOp' and len(v['args']) == 1 and
                        isinstance(v['args'][0], (int, float, numpy.int64,
                                                  numpy.float32, numpy.float64))):
                    if v['op'] == 'Sub':
                        node['args'][i] = -v['args'][0]

    def depart(self, node, info):
        """
        Visits a node.

        @param      node        visited node
        @param      info        info extracted by the visitor
        """
        if 'type' not in info:
            return

        kind = info['type']
        if kind == "arg":
            return
        if kind == "arguments":
            _, buf = self._get_last('FunctionDef')
            for child in info['children']:
                if child['type'] == 'Str':
                    buf['default'].append(child['str'])
                elif child['type'] in ('Num', 'Cst'):
                    buf['default'].append(child['n'])
                elif child['type'] == 'arg':
                    buf['args'].append(
                        (child['str'], child.get('annotation', None)))
                else:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to interpret type '{}' in function definition."
                        "\n{}".format(
                            child['type'], pprint.pformat(info)))
            return

        if kind == "Name":
            op, buf = self._get_last(
                ('Assign', 'BinOp', 'Call', 'Return', 'FunctionDef', 'keyword',
                 'UnaryOp'),
                info)
            if op == 'Assign':
                buf['name'] = info['str']
                return
            elif op in ('BinOp', 'Call'):
                buf['args'].append(info['str'])
                return
            elif op == 'Return':
                buf['code'] = info['str']
                return
            elif op == 'keyword':
                buf['value'] = info['str']
                return
            elif op == 'UnaryOp':
                buf['args'].append(info['str'])
                return
            elif op == 'FunctionDef':
                raise RuntimeError("Default value must be constant, variable '{}' was "
                                   "detected.".format(info['str']))

        if kind in OnnxTranslator._binary_operators:
            _, buf = self._get_last(('BinOp', 'UnaryOp'))
            return
        if kind in ('Call', 'BinOp', 'Assign', 'Return', 'UnaryOp'):
            op, buf = self._get_last(
                ('Call', 'BinOp', 'Assign', 'Return', 'UnaryOp'))
            self._post_process(op, buf)
            self._stack.pop()
            opp, parent = self._get_last(
                ('Call', 'BinOp', 'Assign', 'FunctionDef', 'Return', 'UnaryOp'))
            if opp in ('FunctionDef', 'Return'):
                parent['code'].append((op, buf))
            else:
                parent['args'].append((op, buf))
            self._post_process(None, parent)
            return
        if kind == 'FunctionDef':
            if len(self._stack) == 1:
                self._code_fct = self._stack[-1]
                self._fix_default_values(self._code_fct)
                self._stack = []
                return
        if kind == 'Module':
            return
        if kind == 'Attribute':
            op, buf = self._get_last(('Call', 'BinOp'))

            if len(info["children"]) > 0:
                fir = info["children"][0]
                if fir["type"] == "Name":
                    parent = fir["node"].id
                    info["str"] = f"{parent}.{info['str']}"
                    info["children"][0]["remove"] = True

            buf['name'] = info["str"]
            buf['args'][0] = info["str"]
            return
        if kind in ('Num', 'Cst'):
            op, buf = self._get_last(
                ('List', 'BinOp', 'UnaryOp', 'FunctionDef', 'Call'))
            if op == 'FunctionDef':
                return
            if op == 'List':
                buf['elts'].append(info['n'])
            else:
                buf['args'].append(info['n'])
            return
        if kind == 'Str':
            _, buf = self._get_last('keyword')
            buf['value'] = info['str']
            return
        if kind == 'List':
            op, buf = self._get_last('List')
            value = buf['elts']
            self._post_process(op, buf)
            self._stack.pop()
            opp, parent = self._get_last('keyword')
            parent['value'] = value
            self._post_process(None, parent)
            return
        if kind == 'keyword':
            op, buf = self._get_last('keyword')
            name = buf["name"]
            if 'value' not in buf:
                raise RuntimeError(str(buf))  # pragma: no cover
            value = buf['value']
            self._post_process(op, buf)
            self._stack.pop()
            opp, parent = self._get_last('Call')
            parent['args'].append(OnnxTranslator.Parameter(name, value))
            self._post_process(None, parent)
            return

        raise NotImplementedError(  # pragma: no cover
            "Unable to interpret kind '{}' at {}\n{}\n---\n{}".format(
                info.get('type', '?'), self.make_msg(
                    node), pprint.pformat(info),
                pprint.pformat(self._stack)))



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_onnxruntime*.
"""
import numpy
import onnx.defs
from onnx.helper import make_tensor
import skl2onnx.algebra.onnx_ops as alg
try:
    import skl2onnx.algebra.custom_ops as alg2
except ImportError:  # pragma: no cover
    # older version of skl2onnx
    alg2 = alg
from ...onnx_tools.onnx2py_helper import guess_proto_dtype
from ...onnx_tools.optim.graph_schema_helper import (
    get_defined_inputs, get_defined_outputs, proto2vars)


_schemas = {
    schema.name: schema for schema in onnx.defs.get_all_schemas_with_history()}


class OpRunOnnxEmpty:
    """
    Unique operator for an empty runtime.
    """

    def __init__(self, onnx_node, desc=None, variables=None,
                 dtype=None, **options):
        """
        :param onnx_node: :epkg:`onnx` node
        :param desc: internal representation
        :param variables: registered variables created by previous operators
        :param dtype: float computation type
        :param options: runtime options
        """
        self._provider = 'empty'
        self.onnx_node = onnx_node
        self.desc = desc
        self._schema = _schemas.get(onnx_node.op_type, None)
        if desc is not None:
            if 'atts' in desc:
                for a, b in desc['atts'].items():
                    if not isinstance(b, dict) or 'value' not in b:
                        raise ValueError(  # pragma: no cover
                            f"Unexpected value {b}.")
                    options[a] = b['value']

        self.options = options
        self.dtype = dtype
        self._init(variables)

    def _name_mapping(self, inputs):
        mapping = {}
        new_inputs = []
        for name in inputs:
            if name in mapping:
                i = 0
                new_name = f"{name}_{i}"
                while new_name in mapping:
                    i += 1  # pragma: no cover
                    new_name = f"{name}_{i}"  # pragma: no cover
                mapping[new_name] = name
                new_inputs.append(new_name)
            else:
                new_inputs.append(name)
                mapping[name] = name
        return mapping, new_inputs

    def _guess_proto_type(self, dtype):
        return guess_proto_dtype(dtype)

    def _init(self, variables=None):
        """
        Initializes the node.

        @param      variables               registered variables created by previous operators

        The current implementation for operator *Scan*
        only works for matrices.
        """
        try:
            self.alg_class = getattr(alg2, 'Onnx' + self.onnx_node.op_type)
        except AttributeError:
            try:
                self.alg_class = getattr(alg, 'Onnx' + self.onnx_node.op_type)
            except AttributeError:
                self.alg_class = None
        inputs = list(self.onnx_node.input)
        self.mapping, self.inputs = self._name_mapping(inputs)
        self.outputs = list(self.onnx_node.output)

        options = self.options.copy()
        target_opset = options.pop('target_opset', None)
        domain = options.pop('domain', None)
        # disable_optimisation = options.pop('disable_optimisation', False)
        # ir_version = options.pop('ir_version', None)

        if self.alg_class is None:
            self.onnx_ = self.onnx_node
        elif self.onnx_node.op_type == 'ConstantOfShape':
            for k in options:
                v = options[k]
                if isinstance(v, numpy.ndarray):
                    options[k] = make_tensor(
                        k, self._guess_proto_type(v.dtype),
                        v.shape, v.tolist())

            self.inst_ = self.alg_class(*self.inputs, output_names=self.outputs,
                                        op_version=target_opset, **options)
            inputs = get_defined_inputs(
                self.inputs, variables, dtype=self.dtype)
            try:
                self.onnx_ = self.inst_.to_onnx(inputs, target_opset=target_opset,
                                                domain=domain)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        f"Probable issue as one dimension is null.\n--\n{self.onnx_}")
            except AttributeError as e:  # pragma: no cover
                # older version of skl2onnx
                self.onnx_ = self.inst_.to_onnx(inputs)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(
                        "Probable issue as one dimension is null.\n--\n{}".format(
                            self.onnx_)) from e
        elif self.onnx_node.op_type == 'Scan':
            self.inst_ = self.alg_class(
                *self.inputs, output_names=self.outputs,
                op_version=target_opset, **options)
            inputs = get_defined_inputs(
                self.inputs, variables, dtype=self.dtype)
            outputs = get_defined_outputs(
                self.outputs, self.onnx_node, inputs, variables,
                dtype=self.dtype)
            inputs = [(name, cl.__class__([None, None]))
                      for (name, cl) in inputs]
            outputs = [(name, cl.__class__([None, None]))
                       for (name, cl) in outputs]
            self.onnx_ = self.inst_.to_onnx(inputs, outputs=outputs,
                                            target_opset=target_opset,
                                            domain=domain)
            if "dim_value: 0" in str(self.onnx_):
                raise RuntimeError(  # pragma: no cover
                    f"Probable issue as one dimension is null.\n--\n{self.onnx_}")
        else:
            self.inst_ = self.alg_class(*self.inputs, output_names=self.outputs,
                                        op_version=target_opset, domain=domain,
                                        **options)
            inputs = get_defined_inputs(
                self.inputs, variables, dtype=self.dtype)

            try:
                self.onnx_ = self.inst_.to_onnx(
                    inputs, target_opset=target_opset, domain=domain)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        "Probable issue as one dimension is null.\n--\n{}\n---\n{}".format(
                            self.onnx_, inputs))
            except (RuntimeError, ValueError):  # pragma: no cover
                # Let's try again by forcing output types.
                outputs = get_defined_outputs(
                    self.outputs, self.onnx_node, inputs, variables,
                    dtype=self.dtype)
                self.onnx_ = self.inst_.to_onnx(inputs, outputs=outputs,
                                                target_opset=target_opset,
                                                domain=domain)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        "Probable issue as one dimension is null.\n--\n{}".format(
                            self.onnx_)) from e

        if hasattr(self.onnx_, 'graph'):
            if len(self.onnx_.graph.output) != len(self.outputs):  # pragma: no cover
                # Something is wrong, falls back to default plan.
                outputs = get_defined_outputs(
                    self.outputs, self.onnx_node, inputs, variables,
                    dtype=self.dtype)
                self.onnx_ = self.inst_.to_onnx(inputs, outputs=outputs,
                                                target_opset=target_opset,
                                                domain=domain)
                if "dim_value: 0" in str(self.onnx_):
                    raise RuntimeError(  # pragma: no cover
                        f"Probable issue as one dimension is null.\n--\n{self.onnx_}")
            else:
                lo = list(self.onnx_.graph.output)
                outputs = proto2vars(lo)
        else:
            outputs = [(o, None) for o in self.onnx_.output]

        self.typed_outputs_ = outputs

    def run(self, *args, **kwargs):
        """
        Should be overwritten.
        """
        # inputs = {name: val for name, val in zip(self.inputs, args)}
        raise RuntimeError(  # pragma: no cover
            "This runtime does nothing. Running it is useless.")

    def need_context(self):
        """
        Tells the runtime if this node needs the context
        (all the results produced so far) as it may silently access
        one of them (operator Loop).
        The default answer is `False`.
        """
        return False



# pylint: disable=E0602
"""
@file
@brief Xop helpers.

.. versionadded:: 0.9
"""
from .xop_variable import Variable


def _infer_node_output(node, inputs):
    """
    Infers node outputs for a specific type.

    :param node: :epkg:`NodeProto`
    :param outputs: known inputs
    :return: dtype
    """
    if not isinstance(inputs, dict):
        raise TypeError(  # pragma: no cover
            f"inputs should be OrderedDict not {type(inputs)!r}.")

    if node.op_type == 'Concat':
        type_set = set()
        for v in inputs.values():
            if not isinstance(v, Variable):
                raise TypeError(  # pragma: no cover
                    f"Unexpected type {type(v)!r} for {v!r}.")
            type_set.add(v.dtype)
        if len(type_set) != 1:
            raise RuntimeError(  # pragma: no cover
                f"Unable to guess output type from {type_set!r} (inputs={inputs!r}).")
        dtype = type_set.pop()
        if dtype is None:
            raise RuntimeError(  # pragma: no cover
                f"Guessed output type is None from inputs={inputs!r}.")
        return dtype, [None, None]

    raise NotImplementedError(  # pragma: no cover
        f"Unable to infer type for node type {node.op_type!r} and inputs={inputs!r}.")



"""
@file
@brief Runtime operator.
"""
import numpy


def _get_typed_class_attribute(self, k, atts):
    """
    Converts an attribute into a C++ value.
    """
    ty = atts[k]
    if isinstance(ty, numpy.ndarray):
        v = getattr(self, k)
        return v if v.dtype == ty.dtype else v.astype(ty.dtype)
    if isinstance(ty, bytes):
        return getattr(self, k).decode()
    if isinstance(ty, list):
        v = getattr(self, k)
        if isinstance(v, numpy.ndarray):
            return v
        return [_.decode() for _ in getattr(self, k)]
    if isinstance(ty, int):
        return getattr(self, k)
    raise NotImplementedError(  # pragma: no cover
        f"Unable to convert '{k}' ({getattr(self, k)}).")


def proto2dtype(proto_type):
    """
    Converts a proto type into a :epkg:`numpy` type.

    :param proto_type: example ``onnx.TensorProto.FLOAT``
    :return: :epkg:`numpy` dtype
    """
    from ...onnx_tools.onnx2py_helper import guess_dtype
    return guess_dtype(proto_type)


def dtype_name(dtype):
    """
    Returns the name of a numpy dtype.
    """
    if dtype == numpy.float32:
        return "float32"
    if dtype == numpy.float64:
        return "float64"
    if dtype == numpy.float16:
        return "float16"
    if dtype == numpy.int32:
        return "int32"
    if dtype == numpy.uint32:
        return "uint32"
    if dtype == numpy.int64:
        return "int64"
    if dtype == numpy.int8:
        return "int8"
    if dtype == numpy.uint8:
        return "uint8"
    if dtype == numpy.str_:
        return "str"
    if dtype == numpy.bool_:
        return "bool"
    raise ValueError(
        f"Unexpected dtype {dtype}.")



"""
@file
@brief Tests with onnx backend.
"""
import os
import textwrap
import numpy
from numpy import object as dtype_object
from numpy.testing import assert_almost_equal
import onnx
from onnx.numpy_helper import to_array, to_list
from onnx.backend.test import __file__ as backend_folder


def assert_almost_equal_string(expected, value):
    """
    Compares two arrays knowing they contain strings.
    Raises an exception if the test fails.

    :param expected: expected array
    :param value: value
    """
    def is_float(x):
        try:
            return True
        except ValueError:  # pragma: no cover
            return False

    if all(map(is_float, expected.ravel())):
        expected_float = expected.astype(numpy.float32)
        value_float = value.astype(numpy.float32)
        assert_almost_equal(expected_float, value_float)
    else:
        assert_almost_equal(expected, value)


class OnnxBackendTest:
    """
    Definition of a backend test. It starts with a folder,
    in this folder, one onnx file must be there, then a subfolder
    for each test to run with this model.

    :param folder: test folder
    :param onnx_path: onnx file
    :param onnx_model: loaded onnx file
    :param tests: list of test
    """
    @staticmethod
    def _sort(filenames):
        temp = []
        for f in filenames:
            name = os.path.splitext(f)[0]
            i = name.split('_')[-1]
            temp.append((int(i), f))
        temp.sort()
        return [_[1] for _ in temp]

    @staticmethod
    def _read_proto_from_file(full):
        if not os.path.exists(full):
            raise FileNotFoundError(  # pragma: no cover
                f"File not found: {full!r}.")
        with open(full, 'rb') as f:
            serialized = f.read()
        try:
            loaded = to_array(onnx.load_tensor_from_string(serialized))
        except Exception as e:  # pylint: disable=W0703
            seq = onnx.SequenceProto()
            try:
                seq.ParseFromString(serialized)
                loaded = to_list(seq)
            except Exception:  # pylint: disable=W0703
                try:
                    loaded = onnx.load_model_from_string(serialized)
                except Exception:  # pragma: no cover
                    raise RuntimeError(
                        "Unable to read %r, error is %s, content is %r." % (
                            full, e, serialized[:100])) from e
        return loaded

    @staticmethod
    def _load(folder, names):
        res = []
        for name in names:
            full = os.path.join(folder, name)
            new_tensor = OnnxBackendTest._read_proto_from_file(full)
            if isinstance(new_tensor, (numpy.ndarray, onnx.ModelProto, list)):
                t = new_tensor
            elif isinstance(new_tensor, onnx.TensorProto):
                t = to_array(new_tensor)
            else:
                raise RuntimeError(  # pragma: no cover
                    f"Unexpected type {type(new_tensor)!r} for {full!r}.")
            res.append(t)
        return res

    def __repr__(self):
        "usual"
        return f"{self.__class__.__name__}({self.folder!r})"

    def __init__(self, folder):
        if not os.path.exists(folder):
            raise FileNotFoundError(  # pragma: no cover
                f"Unable to find folder {folder!r}.")
        content = os.listdir(folder)
        onx = [c for c in content if os.path.splitext(c)[-1] in {'.onnx'}]
        if len(onx) != 1:
            raise ValueError(  # pragma: no cover
                f"There is more than one onnx file in {folder!r} ({onx!r}).")
        self.folder = folder
        self.onnx_path = os.path.join(folder, onx[0])
        self.onnx_model = onnx.load(self.onnx_path)

        self.tests = []
        for sub in content:
            full = os.path.join(folder, sub)
            if os.path.isdir(full):
                pb = [c for c in os.listdir(full)
                      if os.path.splitext(c)[-1] in {'.pb'}]
                inputs = OnnxBackendTest._sort(
                    c for c in pb if c.startswith('input_'))
                outputs = OnnxBackendTest._sort(
                    c for c in pb if c.startswith('output_'))

                t = dict(
                    inputs=OnnxBackendTest._load(full, inputs),
                    outputs=OnnxBackendTest._load(full, outputs))
                self.tests.append(t)

    @property
    def name(self):
        "Returns the test name."
        return os.path.split(self.folder)[-1]

    def __len__(self):
        "Returns the number of tests."
        return len(self.tests)

    def _compare_results(self, index, i, e, o, decimal=None):
        """
        Compares the expected output and the output produced
        by the runtime. Raises an exception if not equal.

        :param index: test index
        :param i: output index
        :param e: expected output
        :param o: output
        :param decimal: precision
        """
        if isinstance(e, numpy.ndarray):
            if isinstance(o, numpy.ndarray):
                if decimal is None:
                    if e.dtype == numpy.float32:
                        deci = 6
                    elif e.dtype == numpy.float64:
                        deci = 12
                    else:
                        deci = 7
                else:
                    deci = decimal
                if e.dtype == dtype_object:
                    try:
                        assert_almost_equal_string(e, o)
                    except AssertionError as ex:
                        raise AssertionError(  # pragma: no cover
                            "Output %d of test %d in folder %r failed." % (
                                i, index, self.folder)) from ex
                else:
                    try:
                        assert_almost_equal(e, o, decimal=deci)
                    except AssertionError as ex:
                        raise AssertionError(
                            "Output %d of test %d in folder %r failed." % (
                                i, index, self.folder)) from ex
            elif hasattr(o, 'is_compatible'):
                # A shape
                if e.dtype != o.dtype:
                    raise AssertionError(
                        "Output %d of test %d in folder %r failed "
                        "(e.dtype=%r, o=%r)." % (
                            i, index, self.folder, e.dtype, o))
                if not o.is_compatible(e.shape):
                    raise AssertionError(  # pragma: no cover
                        "Output %d of test %d in folder %r failed "
                        "(e.shape=%r, o=%r)." % (
                            i, index, self.folder, e.shape, o))
        else:
            raise NotImplementedError(
                f"Comparison not implemented for type {type(e)!r}.")

    def is_random(self):
        "Tells if a test is random or not."
        if 'bernoulli' in self.folder:
            return True
        return False

    def run(self, load_fct, run_fct, index=None, decimal=None):
        """
        Executes a tests or all tests if index is None.
        The function crashes if the tests fails.

        :param load_fct: loading function, takes a loaded onnx graph,
            and returns an object
        :param run_fct: running function, takes the result of previous
            function, the inputs, and returns the outputs
        :param index: index of the test to run or all.
        :param decimal: requested precision to compare results
        """
        if index is None:
            for i in range(len(self)):
                self.run(load_fct, run_fct, index=i, decimal=decimal)
            return

        obj = load_fct(self.onnx_model)

        got = run_fct(obj, *self.tests[index]['inputs'])
        expected = self.tests[index]['outputs']
        if len(got) != len(expected):
            raise AssertionError(  # pragma: no cover
                "Unexpected number of output (test %d, folder %r), "
                "got %r, expected %r." % (
                    index, self.folder, len(got), len(expected)))
        for i, (e, o) in enumerate(zip(expected, got)):
            if self.is_random():
                if e.dtype != o.dtype:
                    raise AssertionError(
                        "Output %d of test %d in folder %r failed "
                        "(type mismatch %r != %r)." % (
                            i, index, self.folder, e.dtype, o.dtype))
                if e.shape != o.shape:
                    raise AssertionError(
                        "Output %d of test %d in folder %r failed "
                        "(shape mismatch %r != %r)." % (
                            i, index, self.folder, e.shape, o.shape))
            else:
                self._compare_results(index, i, e, o, decimal=decimal)

    def to_python(self):
        """
        Returns a python code equivalent to the ONNX test.

        :return: code
        """
        from ..onnx_tools.onnx_export import export2onnx
        rows = []
        code = export2onnx(self.onnx_model)
        lines = code.split('\n')
        lines = [line for line in lines
                 if not line.strip().startswith('print') and
                 not line.strip().startswith('# ')]
        rows.append(textwrap.dedent("\n".join(lines)))
        rows.append("oinf = OnnxInference(onnx_model)")
        for test in self.tests:
            rows.append("xs = [")
            for inp in test['inputs']:
                rows.append(textwrap.indent(repr(inp) + ',', '    ' * 2))
            rows.append("]")
            rows.append("ys = [")
            for out in test['outputs']:
                rows.append(textwrap.indent(repr(out) + ',', '    ' * 2))
            rows.append("]")
            rows.append("feeds = {n: x for n, x in zip(oinf.input_names, xs)}")
            rows.append("got = oinf.run(feeds)")
            rows.append("goty = [got[k] for k in oinf.output_names]")
            rows.append("for y, gy in zip(ys, goty):")
            rows.append("    self.assertEqualArray(y, gy)")
            rows.append("")
        code = "\n".join(rows)
        final = "\n".join([f"def {self.name}(self):",
                           textwrap.indent(code, '    ')])
        try:
            from pyquickhelper.pycode.code_helper import remove_extra_spaces_and_pep8
        except ImportError:  # pragma: no cover
            return final
        return remove_extra_spaces_and_pep8(final, aggressive=True)


def enumerate_onnx_tests(series, fct_filter=None):
    """
    Collects test from a sub folder of `onnx/backend/test`.
    Works as an enumerator to start processing them
    without waiting or storing too much of them.

    :param series: which subfolder to load, possible values:
        (`'node'`, ...)
    :param fct_filter: function `lambda testname: boolean`
        to load or skip the test, None for all
    :return: list of @see cl OnnxBackendTest
    """
    root = os.path.dirname(backend_folder)
    sub = os.path.join(root, 'data', series)
    if not os.path.exists(sub):
        raise FileNotFoundError(
            "Unable to find series of tests in %r, subfolders:\n%s" % (
                root, "\n".join(os.listdir(root))))
    tests = os.listdir(sub)
    for t in tests:
        if fct_filter is not None and not fct_filter(t):
            continue
        folder = os.path.join(sub, t)
        content = os.listdir(folder)
        onx = [c for c in content if os.path.splitext(c)[-1] in {'.onnx'}]
        if len(onx) == 1:
            yield OnnxBackendTest(folder)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""

import numpy
from numpy.random import RandomState
from onnx.defs import onnx_opset_version
from ._op import OpRun


def _dropout(X, drop_probability=0.5, seed=0,
             training_mode=False, return_mask=False):
    if drop_probability == 0 or not training_mode:
        if return_mask:
            return X, numpy.ones(X.shape, dtype=bool)
        return (X, )

    rnd = RandomState(seed)
    mask = rnd.uniform(0, 1.0, X.shape) >= drop_probability
    scale = (1. / (1. - drop_probability))
    return (
        (mask * X * scale, mask.astype(bool))
        if return_mask else (mask * X * scale, ))


class DropoutBase(OpRun):

    def __init__(self, onnx_node, desc=None, expected_attributes=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)
        self.nb_outputs = len(onnx_node.output)

    def _private_run(self, X, seed=None, ratio=0.5, training_mode=False):  # pylint: disable=W0221
        return _dropout(X, ratio, seed=seed, return_mask=self.nb_outputs == 2,
                        training_mode=training_mode)


class Dropout_7(DropoutBase):

    atts = {'ratio': 0.5}

    def __init__(self, onnx_node, desc=None, **options):
        DropoutBase.__init__(self, onnx_node, desc=desc,
                             expected_attributes=Dropout_7.atts,
                             **options)

    def _run(self, X, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return self._private_run(X, self.ratio)


class Dropout_12(DropoutBase):

    atts = {'seed': 0}

    def __init__(self, onnx_node, desc=None, **options):
        DropoutBase.__init__(self, onnx_node, desc=desc,
                             expected_attributes=Dropout_12.atts,
                             **options)

    def _run(self, *inputs, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        X = inputs[0]
        ratio = 0.5 if len(inputs) <= 1 else inputs[1]
        training_mode = False if len(inputs) <= 2 else inputs[2]
        return self._private_run(X, seed=self.seed, ratio=ratio,
                                 training_mode=training_mode)


if onnx_opset_version() >= 12:
    Dropout = Dropout_12
else:
    Dropout = Dropout_7  # pragma: no cover



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from textwrap import dedent
from ._op import OpRunUnaryNum


def _leaky_relu(x, alpha):
    sign = (x > 0).astype(x.dtype)
    sign -= ((sign - 1) * alpha).astype(x.dtype)
    return x * sign


def _leaky_relu_inplace(x, alpha):
    sign = (x > 0).astype(x.dtype)
    sign -= ((sign - 1) * alpha).astype(x.dtype)
    x *= sign


class LeakyRelu(OpRunUnaryNum):

    atts = {'alpha': 0.01}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=LeakyRelu.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x)
        return (_leaky_relu(x, self.alpha), )

    def _run_inplace(self, x):
        _leaky_relu_inplace(x, self.alpha)
        return (x, )

    def to_python(self, inputs):
        return (dedent(
            """
            import numpy
            def _leaky_relu(x, alpha):
                sign = (x > 0).astype(x.dtype)
                sign -= ((sign - 1) * alpha).astype(x.dtype)
                return x * sign
            """), f"return _leaky_relu({inputs[0]}, alpha)")



"""
@file
@brief Templates to export an ONNX graph in a way it can we created again
with a python script.

.. versionadded:: 0.7
"""
import sys
import os
from textwrap import dedent
try:
    from functools import cache
except ImportError:  # pragma: no cover
    # python 3.8
    from functools import lru_cache as cache


def _private_get_file(name):
    """
    Retrieves one template.
    """
    this = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(this, f"_onnx_export_templates_{name}.tmpl")
    if not os.path.exists(filename):
        raise FileNotFoundError(  # pragma: no cover
            f"Unable to find template {name!r} in folder {this!r}.")
    with open(filename, "r", encoding="utf-8") as f:
        return dedent(f.read())


if sys.version_info[:2] > (3, 6):
    @cache
    def _get_file(name):
        """
        Retrieves one template.
        """
        return _private_get_file(name)
else:  # pragma: no cover
    def _get_file(name):
        """
        Retrieves one template.
        """
        return _private_get_file(name)


def get_onnx_template():
    """
    Template to export :epkg:`ONNX` into :epkg:`onnx` code.
    """
    return _get_file('onnx')


def get_tf2onnx_template():
    """
    Template to export :epkg:`ONNX` into :epkg:`tensorflow-onnx` code.
    """
    return _get_file('tf2onnx')


def get_numpy_template():
    """
    Template to export :epkg:`ONNX` into :epkg:`numpy` code.
    """
    return _get_file('numpy')


def get_xop_template():
    """
    Template to export :epkg:`ONNX` into a code based on XOP API.
    """
    return _get_file('xop')


def get_cpp_template():
    """
    Template to export :epkg:`ONNX` into a C++ code.
    """
    return _get_file('cpp')


def get_python_template():
    """
    Template to export :epkg:`ONNX` into a python code.
    """
    return _get_file('python')



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnaryNum


class LpNormalization(OpRunUnaryNum):

    atts = {'axis': -1, 'p': 2}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=LpNormalization.atts,
                               **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        norm = numpy.power(numpy.power(x, self.p).sum(
            axis=self.axis), 1. / self.p)
        norm = numpy.expand_dims(norm, self.axis)
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x, norm)
        return (x / norm, )

    def _run_inplace(self, x, norm):
        x /= norm
        return (x, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from .op_roi_align_ import RoiAlignFloat, RoiAlignDouble  # pylint: disable=E0611


class RoiAlign(OpRun):

    atts = {'coordinate_transformation_mode': b'half_pixel',
            'mode': b'avg',
            'output_height': 1,
            'output_width': 1,
            'sampling_ratio': 0,
            'spatial_scale': 1.}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=RoiAlign.atts,
                       **options)
        self.rt32_ = None
        self.rt64_ = None

    def _run(self, X, rois, batch_indices, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if X.dtype == numpy.float32:
            if self.rt32_ is None:
                self.rt32_ = RoiAlignFloat()
                self.rt32_.init(
                    self.coordinate_transformation_mode.decode('ascii'),
                    self.mode.decode('ascii'), self.output_height,
                    self.output_width, self.sampling_ratio, self.spatial_scale)
            rt = self.rt32_
        elif X.dtype == numpy.float64:
            if self.rt64_ is None:
                self.rt64_ = RoiAlignDouble()
                self.rt64_.init(
                    self.coordinate_transformation_mode.decode('ascii'),
                    self.mode.decode('ascii'), self.output_height,
                    self.output_width, self.sampling_ratio, self.spatial_scale)
            rt = self.rt64_
        else:
            raise TypeError(
                f"Unexpected type {X.dtype!r} for X.")

        res = rt.compute(X, rois, batch_indices)
        return (res, )



"""
@file
@brief ONNX Backend for @see cl OnnxInference.

::

    import unittest
    from onnx.backend.test import BackendTest
    backend_test = BackendTest(backend, __name__)
    back_test.include('.*add.*')
    globals().update(backend_test.enable_report().test_cases)
    unittest.main()
"""
from io import BytesIO
import unittest
import numpy
from onnx import version, load as onnx_load
from onnx.backend.base import Backend, BackendRep
from ..onnx_tools.model_checker import check_onnx
from .onnx_inference import OnnxInference
from .onnx_micro_runtime import OnnxMicroRuntime
from .onnx_shape_inference import OnnxShapeInference


class _CombineModels:

    def __init__(self, onnx_inference, shape_inference):
        self.onnx_inference = onnx_inference
        self.shape_inference = shape_inference

    @property
    def input_names(self):
        "Returns the input names."
        return self.onnx_inference.input_names

    @property
    def output_names(self):
        "Returns the output names."
        return self.onnx_inference.output_names

    def run(self, inputs, **kwargs):
        "Runs shape inferance and onnx inference."
        shapes = self.shape_inference.run(**kwargs)
        results = self.onnx_inference.run(inputs, **kwargs)
        for k, v in results.items():
            if not shapes[k].is_compatible(v):
                raise RuntimeError(  # pragma: no cover
                    "Incompatible shapes %r and %r for output %r." % (
                        shapes[k], v.shape, k))
        return results


class OnnxInferenceBackendRep(BackendRep):
    """
    Computes the prediction for an ONNX graph
    loaded with @see cl OnnxInference.

    :param session: @see cl OnnxInference
    """

    def __init__(self, session):
        self._session = session

    def run(self, inputs, **kwargs):  # type: (Any, **Any) -> Tuple[Any, ...]
        """
        Computes the prediction. See @see meth OnnxInference.run.
        """
        if isinstance(inputs, list):
            feeds = {}
            for i, inp in enumerate(self._session.input_names):
                feeds[inp] = inputs[i]
        elif isinstance(inputs, dict):
            feeds = inputs
        elif isinstance(inputs, numpy.ndarray):
            names = self._session.input_names
            if len(names) != 1:
                raise RuntimeError(  # pragma: no cover
                    f"Expecting one input not {len(names)}.")
            feeds = {names[0]: inputs}
        else:
            raise TypeError(  # pragma: no cover
                f"Unexpected input type {type(inputs)!r}.")
        outs = self._session.run(feeds)
        output_names = self._session.output_names
        if output_names is None and hasattr(self._session, 'expected_outputs'):
            output_names = [n[0] for n in self._session.expected_outputs]
        if output_names is None:
            raise RuntimeError(  # pragma: no cover
                f"output_names cannot be None for type {type(self._session)!r}.")
        return [outs[name] for name in output_names]


class OnnxInferenceBackend(Backend):
    """
    ONNX backend following the pattern from
    `onnx/backend/base.py
    <https://github.com/onnx/onnx/blob/main/onnx/backend/base.py>`_.
    This backend can be ran through the following code:

    ::

        import unittest
        from contextlib import redirect_stdout, redirect_stderr
        from io import StringIO
        from onnx.backend.test import BackendTest
        import mlprodict.onnxrt.backend_py as backend

        back_test = BackendTest(backend, __name__)
        back_test.exclude('.*_blvc_.*')
        back_test.exclude('.*_densenet_.*')
        back_test.exclude('.*_densenet121_.*')
        back_test.exclude('.*_inception_.*')
        back_test.exclude('.*_resnet50_.*')
        back_test.exclude('.*_shufflenet_.*')
        back_test.exclude('.*_squeezenet_.*')
        back_test.exclude('.*_vgg19_.*')
        back_test.exclude('.*_zfnet512_.*')
        globals().update(back_test.enable_report().test_cases)
        buffer = StringIO()
        print('---------------------------------')

        if True:
            with redirect_stdout(buffer):
                with redirect_stderr(buffer):
                    res = unittest.main(verbosity=2, exit=False)
        else:
            res = unittest.main(verbosity=2, exit=False)

        testsRun = res.result.testsRun
        errors = len(res.result.errors)
        skipped = len(res.result.skipped)
        unexpectedSuccesses = len(res.result.unexpectedSuccesses)
        expectedFailures = len(res.result.expectedFailures)
        print('---------------------------------')
        print("testsRun=%d errors=%d skipped=%d unexpectedSuccesses=%d "
              "expectedFailures=%d" % (
            testsRun, errors, skipped, unexpectedSuccesses,
            expectedFailures))
        ran = testsRun - skipped
        print("ratio=%f" % (1 - errors * 1.0 / ran))
        print('---------------------------------')
        print(buffer.getvalue())
    """

    @classmethod
    def is_compatible(cls, model, device=None, **kwargs):
        """
        Returns whether the model is compatible with the backend.

        :param model: unused
        :param device: None to use the default device or a string (ex: `'CPU'`)
        :return: boolean
        """
        return device is None or device == 'CPU'

    @classmethod
    def is_opset_supported(cls, model):
        """
        Returns whether the opset for the model is supported by the backend.

        :param model: Model whose opsets needed to be verified.
        :return: boolean and error message if opset is not supported.
        """
        return True, ''

    @classmethod
    def supports_device(cls, device):
        """
        Checks whether the backend is compiled with particular
        device support.
        """
        return device == 'CPU'

    @classmethod
    def create_inference_session(cls, model):
        """
        Instantiates an instance of class @see cl OnnxInference.
        This method should be overwritten to change the runtime
        or any other runtime options.
        """
        return OnnxInference(model)

    @classmethod
    def prepare(cls, model, device=None, **kwargs):
        """
        Loads the model and creates @see cl OnnxInference.

        :param model: ModelProto (returned by `onnx.load`),
            string for a filename or bytes for a serialized model
        :param device: requested device for the computation,
            None means the default one which depends on
            the compilation settings
        :param kwargs: see @see cl OnnxInference
        :return: see @see cl OnnxInference
        """
        if isinstance(model, OnnxInferenceBackendRep):
            return model
        if isinstance(model, (OnnxInference, OnnxMicroRuntime,
                              OnnxShapeInference, _CombineModels)):
            return OnnxInferenceBackendRep(model)
        if isinstance(model, (str, bytes)):
            inf = cls.create_inference_session(model)
            return cls.prepare(inf, device, **kwargs)
        else:
            from ..npy.xop_convert import OnnxSubOnnx
            if isinstance(model, OnnxSubOnnx):
                return OnnxInferenceBackendRep(model)

        onnx_version = tuple(map(int, (version.version.split(".")[:3])))
        onnx_supports_serialized_model_check = onnx_version >= (1, 10, 0)
        bin_or_model = (
            model.SerializeToString() if onnx_supports_serialized_model_check
            else model)
        check_onnx(bin_or_model)
        opset_supported, error_message = cls.is_opset_supported(model)
        if not opset_supported:
            raise unittest.SkipTest(error_message)  # pragma: no cover
        binm = bin_or_model
        if not isinstance(binm, (str, bytes)):
            binm = binm.SerializeToString()
        return cls.prepare(binm, device, **kwargs)

    @classmethod
    def run_model(cls, model, inputs, device=None, **kwargs):
        """
        Computes the prediction.

        :param model: see @see cl OnnxInference returned by function *prepare*
        :param inputs: inputs
        :param device: requested device for the computation,
            None means the default one which depends on
            the compilation settings
        :param kwargs: see @see cl OnnxInference
        :return: predictions
        """
        rep = cls.prepare(model, device, **kwargs)
        return rep.run(inputs, **kwargs)

    @classmethod
    def run_node(cls, node, inputs, device=None, outputs_info=None, **kwargs):
        '''
        This method is not implemented as it is much more efficient
        to run a whole model than every node independently.
        '''
        raise NotImplementedError(  # pragma: no cover
            "Unable to run the model node by node.")


class OnnxInferenceBackendPyC(OnnxInferenceBackend):
    """
    Same backend as @see cl OnnxInferenceBackend but runtime
    is `python_compiled`.
    """

    @classmethod
    def create_inference_session(cls, model):
        return OnnxInference(model, runtime='python_compiled')


class OnnxInferenceBackendOrt(OnnxInferenceBackend):
    """
    Same backend as @see cl OnnxInferenceBackend but runtime
    is `onnxruntime1`.
    """

    @classmethod
    def create_inference_session(cls, model):
        return OnnxInference(model, runtime='onnxruntime1')


class OnnxInferenceBackendMicro(OnnxInferenceBackend):
    """
    Same backend as @see cl OnnxInferenceBackend but runtime
    is @see cl OnnxMicroRuntime.
    """

    @classmethod
    def create_inference_session(cls, model):
        if isinstance(model, str):
            with open(model, 'rb') as f:
                content = onnx_load(f)
        elif isinstance(model, bytes):
            content = onnx_load(BytesIO(model))
        else:
            content = model
        return OnnxMicroRuntime(content)


class OnnxInferenceBackendShape(OnnxInferenceBackend):
    """
    Same backend as @see cl OnnxInferenceBackend but runtime
    is @see cl OnnxShapeInference.
    """

    @classmethod
    def create_inference_session(cls, model):
        if isinstance(model, str):
            with open(model, 'rb') as f:
                content = onnx_load(f)
        elif isinstance(model, bytes):
            content = onnx_load(BytesIO(model))
        else:
            content = model
        return _CombineModels(OnnxInference(content),
                              OnnxShapeInference(content))

    @classmethod
    def run_model(cls, model, inputs, device=None, **kwargs):
        """
        Computes the prediction.

        :param model: see @see cl OnnxShapeInference returned by
            function *prepare*
        :param inputs: inputs
        :param device: requested device for the computation,
            None means the default one which depends on
            the compilation settings
        :param kwargs: see @see cl OnnxInference
        :return: predictions
        """
        rep = cls.prepare(model, device, **kwargs)
        shapes = rep.shape_inference.run(**kwargs)
        results = rep.onnx_inference.run(inputs, **kwargs)
        for k, v in results.items():
            if not shapes[k].is_compatible(v):
                raise RuntimeError(  # pragma: no cover
                    "Incompatible shapes %r and %r for output %r." % (
                        shapes[k], v.shape, k))
        return results


class OnnxInferenceBackendPyEval(OnnxInferenceBackend):
    """
    Same backend as @see cl OnnxInferenceBackend but runtime
    is @see cl OnnxShapeInference.
    """

    @classmethod
    def create_inference_session(cls, model):
        from ..npy.xop_convert import OnnxSubOnnx
        if isinstance(model, str):
            with open(model, 'rb') as f:
                content = onnx_load(f)
        elif isinstance(model, bytes):
            content = onnx_load(BytesIO(model))
        else:
            content = model
        return OnnxSubOnnx(content)

    @classmethod
    def run_model(cls, model, inputs, device=None, **kwargs):
        """
        Computes the prediction.

        :param model: see @see cl OnnxShapeInference returned by
            function *prepare*
        :param inputs: inputs
        :param device: requested device for the computation,
            None means the default one which depends on
            the compilation settings
        :param kwargs: see @see cl OnnxInference
        :return: predictions
        """
        rep = cls.prepare(model, device, **kwargs)
        shapes = rep.shape_inference.run(**kwargs)
        results = rep.onnx_inference.run(inputs, **kwargs)
        for k, v in results.items():
            if not shapes[k].is_compatible(v):
                raise RuntimeError(  # pragma: no cover
                    "Incompatible shapes %r and %r for output %r." % (
                        shapes[k], v.shape, k))
        return results



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from collections import OrderedDict
import numpy
from ._op_helper import _get_typed_class_attribute
from ._op import OpRunUnaryNum, RuntimeTypeError
from ._new_ops import OperatorSchema
from .op_svm_regressor_ import (  # pylint: disable=E0611,E0401
    RuntimeSVMRegressorFloat,
    RuntimeSVMRegressorDouble,
)


class SVMRegressorCommon(OpRunUnaryNum):

    def __init__(self, dtype, onnx_node, desc=None,
                 expected_attributes=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=expected_attributes,
                               **options)
        self._init(dtype=dtype)

    def _get_typed_attributes(self, k):
        return _get_typed_class_attribute(self, k, self.__class__.atts)

    def _find_custom_operator_schema(self, op_name):
        """
        Finds a custom operator defined by this runtime.
        """
        if op_name == "SVMRegressorDouble":
            return SVMRegressorDoubleSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _init(self, dtype):
        if dtype == numpy.float32:
            self.rt_ = RuntimeSVMRegressorFloat(50)
        elif dtype == numpy.float64:
            self.rt_ = RuntimeSVMRegressorDouble(50)
        else:
            raise RuntimeTypeError(  # pragma: no cover
                f"Unsupported dtype={dtype}.")
        atts = [self._get_typed_attributes(k)
                for k in SVMRegressor.atts]
        self.rt_.init(*atts)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        This is a C++ implementation coming from
        :epkg:`onnxruntime`.
        `svm_regressor.cc
        <https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/ml/svm_regressor.cc>`_.
        See class :class:`RuntimeSVMRegressor
        <mlprodict.onnxrt.ops_cpu.op_svm_regressor_.RuntimeSVMRegressor>`.
        """
        pred = self.rt_.compute(x)
        if pred.shape[0] != x.shape[0]:
            pred = pred.reshape(x.shape[0], pred.shape[0] // x.shape[0])
        return (pred, )


class SVMRegressor(SVMRegressorCommon):

    atts = OrderedDict([
        ('coefficients', numpy.empty(0, dtype=numpy.float32)),
        ('kernel_params', numpy.empty(0, dtype=numpy.float32)),
        ('kernel_type', b'NONE'),
        ('n_supports', 0),
        ('one_class', 0),
        ('post_transform', b'NONE'),
        ('rho', numpy.empty(0, dtype=numpy.float32)),
        ('support_vectors', numpy.empty(0, dtype=numpy.float32)),
    ])

    def __init__(self, onnx_node, desc=None, **options):
        SVMRegressorCommon.__init__(
            self, numpy.float32, onnx_node, desc=desc,
            expected_attributes=SVMRegressor.atts,
            **options)


class SVMRegressorDouble(SVMRegressorCommon):

    atts = OrderedDict([
        ('coefficients', numpy.empty(0, dtype=numpy.float64)),
        ('kernel_params', numpy.empty(0, dtype=numpy.float64)),
        ('kernel_type', b'NONE'),
        ('n_supports', 0),
        ('one_class', 0),
        ('post_transform', b'NONE'),
        ('rho', numpy.empty(0, dtype=numpy.float64)),
        ('support_vectors', numpy.empty(0, dtype=numpy.float64)),
    ])

    def __init__(self, onnx_node, desc=None, **options):
        SVMRegressorCommon.__init__(
            self, numpy.float64, onnx_node, desc=desc,
            expected_attributes=SVMRegressorDouble.atts,
            **options)


class SVMRegressorDoubleSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl SVMRegressorDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'SVMRegressorDouble')
        self.attributes = SVMRegressorDouble.atts



"""
@file
@brief Functions to show shortened options in :epkg:`asv` benchmarks.
"""


def expand_onnx_options(model, optim):
    """
    Expands shortened options. Long names hide some part
    of graphs in :epkg:`asv` benchmark. This trick converts
    a string into real conversions options.

    @param      model       model class (:epkg:`scikit-learn`)
    @param      optim       option
    @return                 expanded options

    It is the reverse of function @see fn shorten_onnx_options.
    The following options are handled:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from sklearn.linear_model import LogisticRegression
        from mlprodict.tools.asv_options_helper import expand_onnx_options

        for name in ['cdist', 'nozipmap', 'raw_scores']:
            print(name, ':', expand_onnx_options(LogisticRegression, name))
    """
    if optim == 'cdist':
        options = {model.__class__: {'optim': 'cdist'}}
    elif optim == 'nozipmap':
        options = {model.__class__: {'zipmap': False}}
    elif optim == 'raw_scores':
        options = {model.__class__: {'raw_scores': True, 'zipmap': False}}
    else:
        options = optim  # pragma: no cover
    return options


def shorten_onnx_options(model, opts):
    """
    Shortens onnx options into a string.
    Long names hide some part
    of graphs in :epkg:`asv` benchmark.

    @param      model       model class (:epkg:`scikit-learn`)
    @param      opts        options
    @return                 shortened options

    It is the reverse of function @see fn expand_onnx_options.
    """
    if opts is None:
        return opts
    if opts == {model: {'optim': 'cdist'}}:
        return 'cdist'
    if opts == {model: {'zipmap': False}}:
        return 'nozipmap'
    if opts == {model: {'raw_scores': True, 'zipmap': False}}:
        return 'raw_scores'
    return None


def display_onnx(model_onnx, max_length=1000):
    """
    Returns a shortened string of the model.

    @param      model_onnx      onnx model
    @param      max_length      maximal string length
    @return                     string
    """
    res = str(model_onnx)
    if max_length is None or len(res) <= max_length:
        return res
    begin = res[:max_length // 2]
    end = res[-max_length // 2:]
    return "\n".join([begin, '[...]', end])


def version2number(vers):
    """
    Converts a version number into a real number.
    """
    spl = vers.split('.')
    r = 0
    for i, s in enumerate(spl):
        try:
            vi = int(s)
        except ValueError:
            vi = 0
        r += vi * 10 ** (-i * 3)
    return r



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_cpu*.
"""
import pprint
import numpy
import onnx
import onnx.defs
from onnx import GraphProto
from ._new_ops import OperatorSchema


def _build_schemas():
    res = {}
    for schema in onnx.defs.get_all_schemas_with_history():
        # Multiple version can coexist. The last one is kept.
        if schema.name in res:
            if schema.since_version > res[schema.name].since_version:
                # We keep the most recent one.
                res[schema.name] = schema
        else:
            res[schema.name] = schema
        res[schema.name + '_' + str(schema.since_version)] = schema
    return res


_schemas = _build_schemas()
_at_least_one = {'Constant'}


class RuntimeTypeError(RuntimeError):
    """
    Raised when a type of a variable is unexpected.
    """
    pass


class DefaultNone:
    """
    Default value for parameters when the parameter is not set
    but the operator has a default behaviour for it.
    """
    pass


class RefAttrName:
    """
    Implements a link between a parameter of a function
    and an attribute in node.

    :param name: name of the input
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        "usual"
        return f"{self.__class__.__name__}({self.name!r})"


class OpRun:
    """
    Ancestor to all operators in this subfolder.
    The runtime for every node can checked into
    `ONNX unit tests
    <https://github.com/onnx/onnx/tree/master/onnx/backend/test/case/node>`_.

    :param onnx_node: :epkg:`onnx` node
    :param desc: internal representation
    :param expected_attributes: expected attributes for this node
    :param options: runtime options
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        self._provider = 'python'
        self.onnx_node = onnx_node
        self.desc = desc
        self.inplaces = {}

        if onnx_node.op_type in _schemas:
            self._schema = _schemas[onnx_node.op_type]
        else:
            self._schema = self._find_custom_operator_schema(onnx_node.op_type)
        if self._schema is None:
            raise RuntimeError(  # pragma: no cover
                "Unable to find class name '{}' in available schemas:"
                "(onnx.__version__='{}')\n{}".format(
                    self.__class__.__name__,
                    onnx.__version__,
                    "\n".join(sorted(_schemas))))

        if desc is not None:
            if 'atts' in desc:
                for a, b in desc['atts'].items():
                    if not isinstance(b, dict) or (
                            'value' not in b and 'ref_attr_name' not in b):
                        raise ValueError(  # pragma: no cover
                            f"Unexpected value {b}.")
                    if 'ref_attr_name' in b:
                        options[a] = RefAttrName(b['ref_attr_name'])
                    else:
                        options[a] = (b['value_rt'] if 'value_rt' in b
                                      else b['value'])
        if expected_attributes is not None:
            if onnx_node.op_type in _at_least_one:
                done = 0
                for a, b in expected_attributes.items():
                    if a in options:
                        setattr(self, a, b)
                        done += 1
                if done == 0:
                    raise RuntimeError(  # pragma: no cover
                        "All parameters '{}' are missing from operator '{}', "
                        "given {}.".format(
                            a, onnx_node.op_type, list(sorted(options))))
            else:
                for a, b in expected_attributes.items():
                    if a not in options:
                        if b is DefaultNone:
                            setattr(self, a, None)
                        elif b is None:
                            raise RuntimeError(  # pragma: no cover
                                "Parameter '{}' is missing from operator '{}' "
                                "(class='{}'), given {}.".format(
                                    a, onnx_node.op_type,
                                    self.__class__.__name__,
                                    list(sorted(options))))
                        else:
                            setattr(self, a, b)
        for k, v in options.items():
            setattr(self, k, v)

        if onnx_node.op_type not in _at_least_one:
            for k, v in self._schema.attributes.items():
                if not hasattr(self, k) and getattr(v, 'required', True):
                    raise RuntimeError(  # pragma: no cover
                        "Attribute '{}' is expected based on ONNX specifications "
                        "for node '{}' and options {}.".format(
                            k, onnx_node.op_type, pprint.pformat(options)))

    @staticmethod
    def local_inputs(graph):
        """
        Returns all varibles not registered as inputs and not produced by
        an node inside the graph. This inputs are part of the context
        existing in the graph calling this one.
        """
        if not isinstance(graph, GraphProto):
            raise TypeError(
                f"Unexpected type {type(graph)!r}.")
        local = set()
        known = set()
        for init in graph.initializer:
            known.add(init.name)
        for init in graph.input:
            known.add(init.name)
        for node in graph.node:
            for o in node.output:
                known.add(o)
            for i in node.input:
                if i not in known:
                    local.add(i)
        return list(local)

    def need_context(self):
        """
        Tells the runtime if this node needs the context
        (all the results produced so far) as it may silently access
        one of them (operator Loop).
        The default answer is `False`.
        """
        return False

    def _find_custom_operator_schema(self, op_name):
        raise NotImplementedError(  # pragma: no cover
            f"This method should be overwritten for operator '{op_name}'.")

    def __str__(self):
        """
        usual
        """
        atts = [self.__class__.__name__ + '(',
                f"    op_type={self.onnx_node.op_type}"]
        for k, v in sorted(self.__dict__.items()):
            if k in {'desc', 'onnx_node'}:
                continue
            if 'a' <= k[0] <= 'z' and k[-1] != '_':
                atts.append(f'    {k}={v},')
        atts.append(')')
        return "\n".join(atts)

    def _run(self, *args, **kwargs):
        """
        Should be overwritten.
        """
        raise NotImplementedError(  # pragma: no cover
            "Method '_run' or 'to_python' should be overwritten for operator %s."
            "" % self.__class__.__name__)

    def run(self, *args, **kwargs):  # pylint: disable=E0202
        """
        Calls method ``_run``.
        """
        try:
            res = self._run(*args, **kwargs)
        except TypeError as e:
            raise TypeError(  # pragma: no cover
                "Issues with types {} (operator {}).".format(
                    ", ".join(str(type(_)) for _ in args),
                    self.__class__.__name__)) from e
        except AttributeError as e:
            raise AttributeError(  # pragma: no cover
                "Issues with types {} (operator {}).".format(
                    ", ".join(str(type(_)) for _ in args),
                    self.__class__.__name__)) from e
        return res

    def switch_initializers_dtype(self, dtype_in=numpy.float32,
                                  dtype_out=numpy.float64):
        """
        Switches all initializers to ``numpy.float64``. If *model*
        is None, a simple cast is done.

        @param      dtype_in    previous type
        @param      dtype_out   next type
        @return                 done operations
        """
        done = []
        for k, v in sorted(self.__dict__.items()):
            if k in {'desc', 'onnx_node'}:
                continue
            if isinstance(v, numpy.ndarray):
                if v.dtype == dtype_in:
                    v = v.astype(dtype_out)
                    setattr(self, k, v)
                    done.append(("+", "att", k, getattr(self, k)))
                else:
                    done.append(("-", "att", k, getattr(self, k)))
        if hasattr(self, '_run_no_checks_') and hasattr(self, 'run'):
            self.run = self._run_no_checks_  # pylint: disable=E0202,E1101
        return done

    def enable_inplace_compute(self, index):
        """
        Tells the node that one input can be overwritten.

        @param      index       input index
        """
        self.inplaces[index] = True

    @property
    def args_default(self):
        """
        Returns the list of arguments as well as
        the list of parameters with the default values
        (close to the signature).
        """
        inps = []
        if hasattr(self, 'atts'):
            for k, v in self.atts.items():  # pylint: disable=E1101
                if isinstance(v, (list, tuple, dict)) and len(v) == 0:
                    v = None
                inps.append(f'{k}={v!r}')
        return inps

    @property
    def args_default_modified(self):
        """
        Returns the list of modified parameters.
        """
        if not hasattr(self, 'atts'):
            return None

        inps = []
        for k, v in self.atts.items():  # pylint: disable=E1101
            val = getattr(self, k, None)
            if isinstance(val, numpy.ndarray) and isinstance(v, list):
                val = list(val)
            try:
                if val != v:
                    inps.append(f'{k}={val!r}')
            except ValueError as e:  # pragma: no cover
                raise ValueError(
                    f"Unexpected value for v={v!r} and val={val!r}.") from e
        return inps

    @property
    def args_optional(self):
        """
        Returns the list of optional arguments.
        """
        inps = []
        if hasattr(self, 'optional_inputs'):
            for k, v in self.optional_inputs.items():  # pylint: disable=E1101
                inps.append(f'{k}={v!r}')
        return inps

    @property
    def args_mandatory(self):
        """
        Returns the list of optional arguments.
        """
        if hasattr(self, 'mandatory_inputs'):
            return self.mandatory_inputs  # pylint:  disable=E1101
        return None

    def to_python(self, inputs):
        """
        Returns a python code equivalent to this operator.

        @param      inputs      inputs name
        @return                 imports, python code, both as strings
        """
        raise NotImplementedError(
            f"Operator '{self.__class__.__name__}' has no equivalent python code.")  # pragma: no cover

    def _to_python_numpy(self, inputs, numpy_name):
        return ("import numpy",
                f"return numpy.{numpy_name}({', '.join(inputs)})")

    @property
    def atts_value(self):
        "Returns all parameters in a dictionary."
        if hasattr(self, 'atts'):
            return {k: getattr(self, k)
                    for k in self.atts}  # pylint: disable=E1101
        return None


class OpRunUnary(OpRun):
    """
    Ancestor to all unary operators in this subfolder.
    Checks that inputs type are the same.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)

    def run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202,W0221
        """
        Calls method ``_run``.
        """
        try:
            res = self._run(x, attributes=attributes,
                            verbose=verbose, fLOG=fLOG)
        except TypeError as e:
            raise TypeError(  # pragma: no cover
                "Issues with types {} (binary operator {}).".format(
                    ", ".join(str(type(_)) for _ in [x]),
                    self.__class__.__name__)) from e
        return res


class OpRunArg(OpRunUnary):
    """
    Ancestor to all unary operators in this subfolder
    and which produces position of extremas (ArgMax, ...).
    Checks that inputs type are the same.
    The class must have attributes *axis*, *keepdim*.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=expected_attributes,
                            **options)
        if not hasattr(self, 'keepdims'):
            raise AttributeError(  # pragma: no cover
                "Attribute 'keepdims' is missing.")
        if not hasattr(self, 'axis'):
            raise AttributeError(  # pragma: no cover
                "Attribute 'axis' is missing.")

    def run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202
        """
        Calls method ``_run``.
        """
        res = OpRunUnary.run(self, x, attributes=attributes,
                             verbose=verbose, fLOG=fLOG)
        if res[0].dtype != numpy.int64:
            raise RuntimeTypeError(  # pragma: no cover
                "Output type mismatch: should be '{}' != output '{}' "
                "(operator '{}')".format(
                    numpy.int64, res[0].dtype, self.__class__.__name__))
        return res

    def _run_no_checks_(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return OpRunUnary.run(self, x, attributes=attributes, verbose=verbose, fLOG=fLOG)


class OpRunUnaryNum(OpRunUnary):
    """
    Ancestor to all unary and numerical operators
    in this subfolder. Checks that inputs type
    are the same.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=expected_attributes,
                            **options)

    def run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202
        """
        Calls method ``_run``.
        """
        res = OpRunUnary.run(self, x, attributes=attributes,
                             verbose=verbose, fLOG=fLOG)
        if len(res) == 0 or res[0] is None:
            return res
        if not isinstance(res[0], list) and res[0].dtype != x.dtype:
            raise RuntimeTypeError(  # pragma: no cover
                "Output type mismatch: input '{}' != output '{}' "
                "(operator '{}')".format(
                    x.dtype, res[0].dtype, self.__class__.__name__))
        return res

    def _run_no_checks_(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return OpRunUnary.run(self, x, attributes=attributes, verbose=verbose, fLOG=fLOG)


class OpRunClassifierProb(OpRunUnary):
    """
    Ancestor to all binary operators in this subfolder.
    Checks that inputs type are the same.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=expected_attributes,
                            **options)

    def run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202
        """
        Calls method ``_run``.
        """
        res = OpRunUnary.run(self, x, attributes=attributes,
                             verbose=verbose, fLOG=fLOG)
        if x.dtype in (numpy.float32, numpy.float64) and res[1].dtype != x.dtype:
            raise RuntimeTypeError(  # pragma: no cover
                "Output type mismatch: {} != {} (operator '{}')".format(
                    x.dtype, res[1].dtype, self.__class__.__name__))
        return res

    @property
    def nb_classes(self):
        """
        Returns the number of expected classes.
        """
        return max(len(getattr(self, 'classlabels_ints', [])),
                   len(getattr(self, 'classlabels_int64s', [])),
                   len(self.classlabels_strings))  # pylint: disable=E1101

    def _run_no_checks_(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return OpRunUnary.run(self, x, attributes=attributes, verbose=verbose, fLOG=fLOG)


class OpRunBinary(OpRun):
    """
    Ancestor to all binary operators in this subfolder.
    Checks that inputs type are the same.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)

    def run(self, x, y, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202,W0221
        """
        Calls method ``_run``.
        """
        if x is None or y is None:
            raise RuntimeError(  # pragma: no cover
                f"x and y have different dtype: {type(x)} != {type(y)} ({type(self)})")
        if x.dtype != y.dtype:
            raise RuntimeTypeError(
                "Input type mismatch: {} != {} (operator '{}', shapes {}, {})".format(
                    x.dtype, y.dtype, self.__class__.__name__,
                    x.shape, y.shape))
        try:
            res = self._run(x, y, attributes=attributes,
                            verbose=verbose, fLOG=fLOG)
        except (TypeError, ValueError) as e:  # pragma: no cover
            raise TypeError(
                "Issues with types {} (binary operator {}).".format(
                    ", ".join(str(type(_)) for _ in [x, y]),
                    self.__class__.__name__)) from e
        return res

    def _run_no_checks_(self, x, y, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        Calls method ``_run``.
        """
        try:
            res = self._run(x, y, attributes=attributes,
                            verbose=verbose, fLOG=fLOG)
        except TypeError as e:  # pragma: no cover
            raise TypeError(
                "Issues with types {} (binary operator {}).".format(
                    ", ".join(str(type(_)) for _ in [x, y]),
                    self.__class__.__name__)) from e
        return res


class OpRunBinaryComparison(OpRunBinary):
    """
    Ancestor to all binary operators in this subfolder
    comparing tensors.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRunBinary.__init__(self, onnx_node, desc=desc,
                             expected_attributes=expected_attributes,
                             **options)


class OpRunBinaryNum(OpRunBinary):
    """
    Ancestor to all binary operators in this subfolder.
    Checks that inputs type are the same.
    """

    def __init__(self, onnx_node, desc=None, expected_attributes=None,
                 **options):
        OpRunBinary.__init__(self, onnx_node, desc=desc,
                             expected_attributes=expected_attributes,
                             **options)

    def run(self, x, y, attributes=None, verbose=0, fLOG=None):  # pylint: disable=E0202
        """
        Calls method ``_run``.
        """
        res = OpRunBinary.run(
            self, x, y, attributes=attributes, verbose=verbose, fLOG=fLOG)
        if res[0].dtype != x.dtype:
            raise RuntimeTypeError(
                "Output type mismatch: {} != {} or {} (operator '{}')"
                " type(x)={} type(y)={}".format(
                    x.dtype, res[0].dtype, y.dtype,
                    self.__class__.__name__, type(x), type(y)))
        return res

    def _run_no_checks_(self, x, y, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        Calls method ``_run``.
        """
        return OpRunBinary._run_no_checks_(
            self, x, y, attributes=attributes, verbose=verbose, fLOG=fLOG)


class OpRunBinaryNumpy(OpRunBinaryNum):
    """
    Implements the inplaces logic.
    *numpy_fct* is a binary numpy function which
    takes two matrices and has a argument *out*
    for inplace operations.
    """

    def __init__(self, numpy_fct, onnx_node, desc=None,
                 expected_attributes=None, **options):
        OpRunBinaryNum.__init__(self, onnx_node, desc=desc,
                                expected_attributes=expected_attributes,
                                **options)
        self.numpy_fct = numpy_fct
        self._cannot_inplace_int = self.numpy_fct in (
            numpy.divide, numpy.true_divide)

    def _run(self, a, b, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if (self._cannot_inplace_int and
                numpy.issubdtype(a.dtype, numpy.integer)):
            return (self.numpy_fct(a, b), )
        if self.inplaces.get(0, False) and a.flags['WRITEABLE'] and a.size >= b.size:
            if len(a.shape) == 1 and b.shape == (1, 1):
                a = a.reshape(1, a.shape[0])
            try:
                self.numpy_fct(a, b, out=a)
                return (a, )
            except (ValueError, TypeError):
                return (self.numpy_fct(a, b), )
        if self.inplaces.get(1, False) and b.flags['WRITEABLE'] and a.size <= b.size:
            if len(b.shape) == 1 and a.shape == (1, 1):
                b = b.reshape(b.shape[0], 1)
            try:
                self.numpy_fct(a, b, out=b)
                return (b, )
            except (ValueError, TypeError):
                return (self.numpy_fct(a, b), )
        return (self.numpy_fct(a, b), )

    def to_python(self, inputs):
        """
        Returns a python code equivalent to this operator.

        @param      inputs      inputs name
        @return                 imports, python code, both as strings
        """
        lines = [
            "# inplaces not take into account {}-{}".format(
                self.inplaces.get(0, False), self.inplaces.get(1, False)),
            f"return numpy.{self.numpy_fct.__name__}({', '.join(inputs)})"
        ]
        return "import numpy", "\n".join(lines)


class OpRunReduceNumpy(OpRunUnaryNum):
    """
    Implements the reduce logic.
    It must have a parameter *axes*.
    """

    def __init__(self, onnx_node, desc=None,
                 expected_attributes=None, **options):
        if ('noop_with_empty_axes' not in expected_attributes and
                'axes' not in expected_attributes):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'axes' is expected but not found in {} "
                "from class {}".format(expected_attributes, type(self)))
        if (expected_attributes.get('noop_with_empty_axes', 0) and
                (expected_attributes['axes'] is None or
                    len(expected_attributes['axes']) == 0)):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'axes' cannot be empty as {} (noop_with_empty_axes=1) "
                "from class {}".format(expected_attributes, type(self)))
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=expected_attributes,
                               **options)
        if isinstance(self.axes, numpy.ndarray):  # pylint: disable=E0203
            if (len(self.axes.shape) == 0 or  # pylint: disable=E0203,E1101
                    self.axes.shape[0] == 0):  # pylint: disable=E0203,E1101
                self.axes = None
            else:
                self.axes = tuple(self.axes)
        elif self.axes in [[], tuple()]:  # pylint: disable=E0203
            self.axes = None
        elif isinstance(self.axes, list):  # pylint: disable=E0203
            self.axes = tuple(self.axes)


class OpRunCustom(OpRun):
    """
    Automates some methods for custom operators defined
    outside *mlprodict*.
    """

    class OpRunCustomSchema(OperatorSchema):
        """
        Custom schema.
        """

        def __init__(self, cls):
            OperatorSchema.__init__(self, cls.__name__)
            self.attributes = cls.atts

    def __init__(self, onnx_node, desc=None,
                 expected_attributes=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)

    def _find_custom_operator_schema(self, op_name):
        """
        Finds a custom operator defined by this runtime.
        """
        if (op_name == self.__class__.__name__ or
                (hasattr(self.__class__, 'op_name') and
                    self.__class__.op_name == op_name)):  # pylint: disable=E1101
            return OpRunCustom.OpRunCustomSchema(self.__class__)
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")



"""
@file
@brief ONNX Backend for @see cl OnnxInference.

::

    import unittest
    from contextlib import redirect_stdout, redirect_stderr
    from io import StringIO
    from onnx.backend.test import BackendTest
    import mlprodict.onnxrt.backend_py as backend

    back_test = BackendTest(backend, __name__)
    back_test.exclude('.*_blvc_.*')
    back_test.exclude('.*_densenet_.*')
    back_test.exclude('.*_densenet121_.*')
    back_test.exclude('.*_inception_.*')
    back_test.exclude('.*_resnet50_.*')
    back_test.exclude('.*_shufflenet_.*')
    back_test.exclude('.*_squeezenet_.*')
    back_test.exclude('.*_vgg19_.*')
    back_test.exclude('.*_zfnet512_.*')
    globals().update(back_test.enable_report().test_cases)
    buffer = StringIO()
    print('---------------------------------')

    if True:
        with redirect_stdout(buffer):
            with redirect_stderr(buffer):
                res = unittest.main(verbosity=2, exit=False)
    else:
        res = unittest.main(verbosity=2, exit=False)

    testsRun = res.result.testsRun
    errors = len(res.result.errors)
    skipped = len(res.result.skipped)
    unexpectedSuccesses = len(res.result.unexpectedSuccesses)
    expectedFailures = len(res.result.expectedFailures)
    print('---------------------------------')
    print("testsRun=%d errors=%d skipped=%d unexpectedSuccesses=%d "
          "expectedFailures=%d" % (
        testsRun, errors, skipped, unexpectedSuccesses,
        expectedFailures))
    ran = testsRun - skipped
    print("ratio=%f" % (1 - errors * 1.0 / ran))
    print('---------------------------------')
    lines = buffer.getvalue().split('\\n')
    print("\\n".join(line for line in lines
          if "skipped 'no matched include pattern'" not in line))
"""
from .backend import OnnxInferenceBackend

is_compatible = OnnxInferenceBackend.is_compatible
prepare = OnnxInferenceBackend.prepare
run = OnnxInferenceBackend.run_model
supports_device = OnnxInferenceBackend.supports_device



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _pad_impl(data, raw_pads, mode, constant_values=0.0):
    input_rank = data.ndim
    if input_rank * 2 != raw_pads.size:
        raise RuntimeError(  # pragma: no cover
            'The number of elements in raw_pads should be 2 * data_rank')

    half = raw_pads.shape[0] // 2
    pad_width = tuple((raw_pads[i], raw_pads[i + half])
                      for i in range(0, half))

    if mode == 'constant':
        return numpy.pad(data, pad_width=pad_width, mode=mode,
                         constant_values=constant_values)
    return numpy.pad(data, pad_width=pad_width, mode=mode)


def onnx_pad(data, pads, constant_value=None, mode='constant'):
    """
    Implements :epkg:`numpy:pad` based on ONNX signature.

    :param data: data to pad
    :param pads: tensor of integers indicating the number of
        padding elements to add or remove (if negative) at the
        beginning and end of each axis. For 2D input tensor, it
        is the number of pixels. `pads` should be a 1D tensor of
        shape `[2 * input_rank]`. `pads` format should be:
        `[x1_begin, x2_begin,...,x1_end, x2_end,...]`, where `xi_begin` is
        the number of pad values added at the beginning of axis `i`
        and xi_end, the number of pad values added at the end of axis `i`.
    :param constant_value: A scalar value to be used if the mode chosen is
        `constant` (by default it is 0, empty string or False).
    :param mode: Supported modes: `constant`(default), `reflect`, `edge`
    :return: tensor after padding
    """
    return _pad_impl(
        data, pads, mode=mode,
        constant_values=constant_value or numpy.array(
            [0], dtype=data.dtype.type))


class Pad(OpRun):

    atts = {'mode': b'constant'}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Pad.atts,
                       **options)
        self.mode_ = self.mode.decode('ascii')

    def _run(self, data, pads, constant_value=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if constant_value is None:
            constant_value = 0
        return (_pad_impl(data, pads, mode=self.mode_,
                          constant_values=constant_value), )



"""
@file
@brief Helpers to compare executions.
"""
import copy
import numpy
from .validate_difference import measure_relative_difference


def _side_by_side_by_values_inputs(sess, inputs, i):
    if isinstance(sess, tuple) and inputs is None:
        new_sess, new_inputs = sess
    elif isinstance(inputs, list):
        new_sess = sess
        new_inputs = inputs[i]
    else:
        new_sess = sess
        new_inputs = copy.deepcopy(inputs)
    return new_sess, new_inputs


def side_by_side_by_values(sessions, *args, inputs=None,
                           return_results=False, **kwargs):
    """
    Compares the execution of two sessions.
    It calls method :meth:`OnnxInference.run
    <mlprodict.onnxrt.onnx_inference.OnnxInference.run>`
    with value ``intermediate=True`` and compares the results.

    :param sessions: list of class @see cl OnnxInference
    :param inputs: inputs
    :param args: additional parameters for
        :meth:`OnnxInference.run
        <mlprodict.onnxrt.onnx_inference.OnnxInference.run`
    :param return_results: if True, returns the results as well.
    :param kwargs: additional parameters for
        :meth:`OnnxInference.run
        <mlprodict.onnxrt.onnx_inference.OnnxInference.run`
    :return: list of dictionaries

    The first session is considered as the baseline.
    See notebook :ref:`onnxsbsrst` for an example.
    If *inputs* is None, the function assumes
    *sessions* is a list of *tuple(sessions, inputs)*
    because sometimes inputs must be different.

    .. versionchanged:: 0.7
        Parameter *return_results* was added. The function
        returns the execution order when available.
    """
    if not kwargs.get('intermediate', True):
        raise ValueError(  # pragma: no cover
            "kwargs must not set intermediate to True")
    kwargs['intermediate'] = True
    verbose = kwargs.get('verbose', 0)
    fLOG = kwargs.get('fLOG', None)

    # run
    results = []
    orders = []
    for i, sess in enumerate(sessions):
        if (hasattr(sess, 'runtime') and hasattr(sess, 'inplace') and
                sess.runtime in (None, 'python') and sess.inplace):
            raise ValueError(
                "You must disable the inplace mechanism in order to get "
                "true results. See OnnxInference constructor.")
        new_sess, new_inputs = _side_by_side_by_values_inputs(sess, inputs, i)
        if verbose > 0 and fLOG:
            fLOG(  # pragma: no cover
                f'[side_by_side_by_values] run session {i + 1}/{len(sessions)}')
        res = new_sess.run(new_inputs, *args, **kwargs)
        order = new_sess.get_execution_order()
        results.append([(k, v) for k, v in res.items()])
        orders.append(order)

    # same number of results?
    rows = []
    row = {"metric": "nb_results", 'step': -1}
    for i, res in enumerate(results):
        row["v[%d]" % i] = len(res)
    mnd = min(map(len, results))
    mxd = max(map(len, results))
    row['cmp'] = 'OK' if mnd == mxd else '!='
    rows.append(row)

    merged = merge_results(results)

    # analysis
    for i in range(len(merged)):  # pylint: disable=C0200
        for metric in ('rel-diff', 'abs-diff'):
            row = {'step': i}
            name, res_row = merged[i]
            row['name'] = name
            row['metric'] = metric

            vals = []
            for j, r in enumerate(res_row):
                order = orders[j]
                if order is not None:
                    row['order[%d]' % j] = order.get(
                        ('res', name), (numpy.nan, ))[0]
                row['value[%d]' % j] = r
                if hasattr(r, 'shape'):
                    row['shape[%d]' % j] = r.shape

                if j == 0:
                    row['v[%d]' % j] = 0
                elif res_row[0] is not None and r is not None:
                    v = measure_relative_difference(
                        res_row[0], r, abs_diff=metric == 'abs-diff')
                    row['v[%d]' % j] = v
                    vals.append(v)

            if len(vals) > 0:
                diff = max(vals)
                if diff < 1e-5:
                    row['cmp'] = 'OK'
                elif diff < 0.0001:  # pragma: no cover
                    row['cmp'] = 'e<0.0001'
                elif diff < 0.001:  # pragma: no cover
                    row['cmp'] = 'e<0.001'
                elif diff < 0.01:  # pragma: no cover
                    row['cmp'] = 'e<0.01'
                elif diff < 0.1:  # pragma: no cover
                    row['cmp'] = 'e<0.1'
                else:  # pragma: no cover
                    row['cmp'] = f"ERROR->={diff:1.1f}"

        rows.append(row)
    if return_results:
        return rows, results
    return rows


def merge_results(results):
    """
    Merges results by name. The first ones
    are used to keep the order.

    :param results: results of intermediate variables
    :return: list of tuple
    """
    # matrix of names
    rows = [(k, []) for k, _ in results[0]]
    positions = {k[0]: i for i, k in enumerate(rows)}
    todos = []
    for result in results:
        todo = []
        for row in rows:
            row[1].append(None)
        for i, (k, v) in enumerate(result):
            pos = positions.get(k, None)
            if pos is None:
                todo.append((i, k, v))
            else:
                rows[pos][1][-1] = (v, i)
        todos.append(todo)

    # left over
    if len(todos) > 0:
        for i, todo in enumerate(todos):
            if len(todo) == 0:
                continue
            for pos, name, val in todo:
                pos1 = pos + 1
                found = -1
                for ik, row in enumerate(rows):
                    if row[1][i] is not None and row[1][i][1] == pos1:
                        found = ik
                        break
                vv = [None] * len(results)
                if found == -1:
                    vv[i] = (val, len(rows))
                    rows.append((name, vv))
                else:
                    vv[i] = (val, pos)
                    rows.insert(found, (name, vv))

    # final
    final = []
    for row in rows:
        nrow = (row[0], [_ if _ is None else _[0] for _ in row[1]])
        final.append(nrow)
    return final
